@@ -17,15 +17,17 @@ let fatal msg =
   let () = Format.eprintf "Error: %s@\n%!" msg in
   exit 1
 
-let dummy_loc = { loc_start = Lexing.dummy_pos; loc_end = Lexing.dummy_pos }
-let mk_code s = { code = s; loc = dummy_loc }
+let mk_code s = { code = s; loc = None }
 
 let print_code fmt c =
-  let loc = c.loc.loc_start in
-  (* Print the line location as a source annotation *)
-  let padding = String.make (loc.pos_cnum - loc.pos_bol + 1) ' ' in
-  let code_insert = asprintf "\n# %i \"%s\"\n%s%s" loc.pos_lnum loc.pos_fname padding c.code in
-  fprintf fmt "@[@<0>%s@]@\n" code_insert
+  match c.loc with
+  | None -> fprintf fmt "%s" c.code
+  | Some loc ->
+    (* Print the line location as a source annotation *)
+    let loc = loc.loc_start in
+    let padding = String.make (loc.pos_cnum - loc.pos_bol + 1) ' ' in
+    let code_insert = asprintf "\n# %i \"%s\"\n%s%s" loc.pos_lnum loc.pos_fname padding c.code in
+    fprintf fmt "@[@<0>%s@]@\n" code_insert
 
 module StringSet = Set.Make(String)
 
@@ -41,7 +43,27 @@ let string_split s =
   in
   if len == 0 then [] else split 0
 
-let plugin_name = "__coq_plugin_name"
+(* Used to generate unique ids *)
+let file_name = ref ""
+
+(* We used to output __plugin_name = "bla" then refer to __plugin_name in generated code
+   but this is not robust to eg
+
+   DECLARE PLUGIN "bla"
+   open G_ltac
+   use the plugin name
+
+   since g_ltac has no mli its __plugin_name will shadow the local one and wreak havoc
+*)
+let plugin_name = ref None
+
+let force_is_plugin ~what () = match !plugin_name with
+  | Some (Some n) -> n
+  | Some None | None -> fatal ("DECLARE PLUGIN required before "^what)
+
+let check_is_plugin ~what () = match !plugin_name with
+  | Some b -> b
+  | None -> fatal ("DECLARE PLUGIN required before "^what)
 
 let print_list fmt pr l =
   let rec prl fmt = function
@@ -267,13 +289,25 @@ let print_rule fmt r =
   let pr_prd fmt prd = print_list fmt print_prod prd in
   fprintf fmt "@[(%a,@ %a,@ %a)@]" pr_lvl r.grule_label pr_asc r.grule_assoc pr_prd (List.rev r.grule_prods)
 
+let gramext_count = ref 0
+
+let gramext_plugin_uid name =
+  let cnt = !gramext_count in
+  incr gramext_count;
+  " ~plugin_uid:(\""^name^"\", \""^(!file_name)^":"^string_of_int cnt^"\")"
+
+let grammar_extend () =
+  match check_is_plugin ~what:"GRAMMAR EXTEND" () with
+  | Some name -> "Egramml.grammar_extend"^gramext_plugin_uid name
+  | None -> "Pcoq.grammar_extend"
+
 let print_entry fmt e = match e.gentry_rules with
 | GDataReuse (pos, r) ->
   let rules = List.rev r in
   let pr_pos fmt pos = print_opt fmt print_string pos in
   let pr_prd fmt prd = print_list fmt print_prod prd in
-  fprintf fmt "let () =@ @[Pcoq.grammar_extend@ %s@ @[(Pcoq.Reuse (%a, %a))@]@]@ in@ "
-    e.gentry_name pr_pos pos pr_prd rules
+  fprintf fmt "let () =@ @[%s@ %s@ @[(Pcoq.Reuse (%a, %a))@]@]@ in@ "
+    (grammar_extend ()) e.gentry_name pr_pos pos pr_prd rules
 | GDataFresh (pos, rules) ->
   let print_rules fmt rules = print_list fmt print_rule rules in
   let pr_check fmt () = match pos with
@@ -281,8 +315,8 @@ let print_entry fmt e = match e.gentry_rules with
   | Some _ -> fprintf fmt ""
   in
   let pos = match pos with None -> First | Some pos -> pos in
-  fprintf fmt "%alet () =@ @[Pcoq.grammar_extend@ %s@ @[(Pcoq.Fresh@ (%a, %a))@]@]@ in@ "
-    pr_check () e.gentry_name print_position pos print_rules rules
+  fprintf fmt "%alet () =@ @[%s@ %s@ @[(Pcoq.Fresh@ (%a, %a))@]@]@ in@ "
+    pr_check () (grammar_extend ()) e.gentry_name print_position pos print_rules rules
 
 let print_ast fmt ext =
   let () = fprintf fmt "let _ = @[" in
@@ -342,37 +376,56 @@ let print_atts_right fmt = function
   | Some atts ->
     let rec aux fmt = function
       | [] -> assert false
-      | [_,y] -> fprintf fmt "%s" y
-      | (_,y) :: rem -> fprintf fmt "(%s ++ %a)" y aux rem
+      | [_,y] -> print_code fmt y
+      | (_,y) :: rem -> fprintf fmt "(%a ++ %a)" print_code y aux rem
     in
     let nota = match atts with [_] -> "" | _ -> "Attributes.Notations." in
     fprintf fmt "(Attributes.parse %s%a atts)" nota aux atts
 
 let understand_state = function
-  | "close_proof" -> "vtcloseproof", false
-  | "open_proof" -> "vtopenproof", true
-  | "proof" -> "vtmodifyproof", false
-  | "proof_opt_query" -> "vtreadproofopt", false
-  | "proof_query" -> "vtreadproof", false
-  | "read_program" -> "vtreadprogram", false
-  | "program" -> "vtmodifyprogram", false
-  | "declare_program" -> "vtdeclareprogram", false
-  | "program_interactive" -> "vtopenproofprogram", false
+  | "close_proof" -> "vtcloseproof", ["lemma"; "pm"]
+  | "open_proof" -> "vtopenproof", []
+  | "proof" -> "vtmodifyproof", ["pstate"]
+  | "proof_opt_query" -> "vtreadproofopt", ["pstate"]
+  | "proof_query" -> "vtreadproof", ["pstate"]
+  | "read_program" -> "vtreadprogram", ["pm"]
+  | "program" -> "vtmodifyprogram", ["pm"]
+  | "declare_program" -> "vtdeclareprogram", ["pm"]
+  | "program_interactive" -> "vtopenproofprogram", ["pm"]
+  | "opaque_access" -> "vtopaqueaccess", ["opaque_access"]
   | s -> fatal ("unsupported state specifier: " ^ s)
 
+let rec pr_named_arguments fmt = function
+| [] -> assert false
+| [s] -> fprintf fmt "~%s" s
+| s :: l -> fprintf fmt "~%s@ %a" s pr_named_arguments l
+
+let pr_begin_wrapper fmt = function
+| [] -> fprintf fmt "fun () ->"
+| args -> fprintf fmt "fun %a ->" pr_named_arguments args
+
+let pr_end_wrapper fmt = function
+| [] -> fprintf fmt ""
+| args -> fprintf fmt "@ %a" pr_named_arguments args
+
 let print_body_state state fmt r =
   let state = match r.vernac_state with Some _ as s -> s | None -> state in
   match state with
-  | None -> fprintf fmt "Vernacextend.vtdefault (fun () -> %a)" print_code r.vernac_body
+  | None -> fprintf fmt "Vernactypes.vtdefault (fun () -> %a)" print_code r.vernac_body
   | Some "CUSTOM" -> print_code fmt r.vernac_body
   | Some state ->
-    let state, unit_wrap = understand_state state in
-    fprintf fmt "Vernacextend.%s (%s%a)" state (if unit_wrap then "fun () ->" else "")
-      print_code r.vernac_body
+    let state, wrap = understand_state state in
+    fprintf fmt "Vernactypes.%s (%a (%a)%a)" state pr_begin_wrapper wrap
+      print_code r.vernac_body pr_end_wrapper wrap
 
 let print_body_fun state fmt r =
-  fprintf fmt "let coqpp_body %a%a = @[%a@] in "
-    print_binders r.vernac_toks print_atts_left r.vernac_atts (print_body_state state) r
+  match r.vernac_synterp with
+  | None ->
+    fprintf fmt "let coqpp_body %a%a = @[%a@] in "
+      print_binders r.vernac_toks print_atts_left r.vernac_atts (print_body_state state) r
+  | Some (id,pe) ->
+    fprintf fmt "let coqpp_body %a%a = @[(let %s = %a in %a)@] in "
+      print_binders r.vernac_toks print_atts_left r.vernac_atts id print_code pe  (print_body_state state) r
 
 let print_body state fmt r =
   fprintf fmt "@[(%afun %a?loc ~atts ()@ -> coqpp_body %a%a)@]"
@@ -409,7 +462,8 @@ let print_entry fmt = function
 
 let print_ast fmt ext =
   let pr fmt () =
-    fprintf fmt "Vernacextend.vernac_extend ~command:\"%s\" %a ?entry:%a %a"
+    fprintf fmt "Vernacextend.static_vernac_extend ~plugin:%s ~command:\"%s\" %a ?entry:%a %a"
+      (match check_is_plugin ~what:"VERNAC EXTEND" () with | Some name -> "(Some \""^name^"\")" | None -> "None")
       ext.vernacext_name print_classifier ext.vernacext_class
       print_entry ext.vernacext_entry (print_rules ext.vernacext_state) ext.vernacext_rules
   in
@@ -448,8 +502,9 @@ let print_ast fmt ext =
   in
   let pr fmt () =
     let level = match ext.tacext_level with None -> 0 | Some i -> i in
-    fprintf fmt "Tacentries.tactic_extend %s \"%s\" ~level:%i %a%a"
-      plugin_name ext.tacext_name level
+    let name = force_is_plugin ~what:"TACTIC EXTEND" () in
+    fprintf fmt "Tacentries.tactic_extend \"%s\" \"%s\" ~level:%i %a%a"
+      name ext.tacext_name level
       deprecation ext.tacext_deprecated
       print_rules ext.tacext_rules
   in
@@ -470,7 +525,7 @@ struct
 let terminal s =
   let p =
     if s <> "" && s.[0] >= '0' && s.[0] <= '9' then "CLexer.terminal_number"
-    else "CLexer.terminal" in
+    else "Pcoq.terminal" in
   let c = Printf.sprintf "Pcoq.Symbol.token (%s \"%s\")" p s in
   SymbQuote c
 
@@ -517,9 +572,10 @@ let print_printer fmt = function
 let print_ast fmt arg =
   let name = arg.vernacargext_name in
   let pr fmt () =
-    fprintf fmt "Vernacextend.vernac_argument_extend ~name:%a @[{@\n\
+    fprintf fmt "Vernacextend.vernac_argument_extend ~plugin:\"%s\" ~name:%a @[{@\n\
       Vernacextend.arg_parsing = %a;@\n\
       Vernacextend.arg_printer = fun env sigma -> %a;@\n}@]"
+      (force_is_plugin ~what:"VERNAC ARGUMENT EXTEND" ())
       print_string name print_rules (name, arg.vernacargext_rules)
       print_printer arg.vernacargext_printer
   in
@@ -601,13 +657,14 @@ let print_ast fmt arg =
   | None -> default_printer
   in
   let pr fmt () =
-    fprintf fmt "Tacentries.argument_extend ~name:%a @[{@\n\
+    fprintf fmt "Tacentries.argument_extend ~plugin:\"%s\" ~name:%a @[{@\n\
       Tacentries.arg_parsing = %a;@\n\
       Tacentries.arg_tag = @[%a@];@\n\
       Tacentries.arg_intern = @[%a@];@\n\
       Tacentries.arg_subst = @[%a@];@\n\
       Tacentries.arg_interp = @[%a@];@\n\
-      Tacentries.arg_printer = @[((fun env sigma -> %a), (fun env sigma -> %a), (fun env sigma -> %a))@];@\n}@]"
+                 Tacentries.arg_printer = @[((fun env sigma -> %a), (fun env sigma -> %a), (fun env sigma -> %a))@];@\n}@]"
+      (force_is_plugin ~what:"ARGUMENT EXTEND" ())
       print_string name
       VernacArgumentExt.print_rules (name, arg.argext_rules)
       pr_tag arg.argext_type
@@ -619,8 +676,12 @@ let print_ast fmt arg =
 end
 
 let declare_plugin fmt name =
-  fprintf fmt "let %s = \"%s\"@\n" plugin_name name;
-  fprintf fmt "let _ = Mltop.add_known_module %s@\n" plugin_name
+  Option.iter (fprintf fmt "let _ = Mltop.add_known_module \"%s\"@\n") name;
+  let () = match !plugin_name with
+    | None -> plugin_name := Some name
+    | Some _ -> fatal "Multiple DECLARE PLUGIN not allowed";
+  in
+  ()
 
 let pr_ast fmt = function
 | Code s -> fprintf fmt "%a@\n" print_code s
@@ -657,8 +718,9 @@ let () =
   let output = output_name file in
   let ast = parse_file file in
   let chan = open_out output in
+  let () = file_name := Filename.basename file in
   let fmt = formatter_of_out_channel chan in
-  let iter ast = Format.fprintf fmt "@[%a@]%!"pr_ast ast in
+  let iter ast = Format.fprintf fmt "@[%a@]%!" pr_ast ast in
   let () = List.iter iter ast in
   let () = close_out chan in
   exit 0
