@@ -12,6 +12,7 @@ if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
 export COQEXTRAFLAGS='-native-compiler no'
 ( cd "${CI_BUILD_DIR}/equations"
   [ -e Makefile.coq ] || ./configure.sh coq
-  make ci
+  make -f Makefile.coq .merlin
+  make
   make install
 )
