@@ -20,6 +20,7 @@ could imagine extending it.
 
 import os
 import re
+import shlex
 from itertools import chain
 from collections import defaultdict
 
@@ -233,12 +234,11 @@ class CoqObject(ObjectDescription):
             signode['first'] = (not self.names)
             self._record_name(name, targetid, signode)
         else:
-            # todo: make the following a real error or warning
-            # todo: then maybe the above "if" is not needed
-            names_in_subdomain = self.subdomain_data()
-            if name in names_in_subdomain:
-                print("Duplicate", self.subdomain, "name: ", name)
-            # self._warn_if_duplicate_name(names_in_subdomain, name, signode)
+            # We don't warn for duplicates in the SSReflect chapter, because
+            # it's the style of this chapter to repeat all the defined
+            # objects at the end.
+            if self.env.docname != 'proof-engine/ssreflect-proof-language':
+                self._warn_if_duplicate_name(self.subdomain_data(), name, signode)
         return targetid
 
     def _add_index_entry(self, name, target):
@@ -953,6 +953,10 @@ class CoqtopBlocksTransform(Transform):
     def is_coqtop_block(node):
         return isinstance(node, nodes.Element) and 'coqtop_options' in node
 
+    @staticmethod
+    def is_coqtop_args_field(node):
+        return isinstance(node, nodes.field) and node.children[0].rawsource == 'COQTOP_ARGS'
+
     @staticmethod
     def split_lines(source):
         r"""Split Coq input into chunks, which may include single- or
@@ -1094,7 +1098,9 @@ class CoqtopBlocksTransform(Transform):
         """Add coqtop's responses to a Sphinx AST
 
         Finds nodes to process using is_coqtop_block."""
-        with CoqTop(color=True) as repl:
+        arg_fields = self.document.traverse(CoqtopBlocksTransform.is_coqtop_args_field)
+        additional_args = [arg for field in arg_fields for arg in shlex.split(field.children[1].rawsource)]
+        with CoqTop(color=True, args=additional_args) as repl:
             repl.send_initial_options()
             for node in self.document.traverse(CoqtopBlocksTransform.is_coqtop_block):
                 try:
