@@ -18,6 +18,8 @@ open Evd
 open Termops
 open Namegen
 
+module ERelevance = EConstr.ERelevance
+
 module RelDecl = Context.Rel.Declaration
 module NamedDecl = Context.Named.Declaration
 
@@ -34,12 +36,13 @@ let create_clos_infos env sigma flags =
 let finalize ?abort_on_undefined_evars sigma f =
   let sigma = minimize_universes sigma in
   let uvars = ref Univ.Level.Set.empty in
-  let v = f (fun c ->
-      let varsc = EConstr.universes_of_constr sigma c in
-      let c = EConstr.to_constr ?abort_on_undefined_evars sigma c in
-      uvars := Univ.Level.Set.union !uvars varsc;
-      c)
+  let nf_constr c =
+    let _, varsc = EConstr.universes_of_constr sigma c in
+    let c = EConstr.to_constr ?abort_on_undefined_evars sigma c in
+    uvars := Univ.Level.Set.union !uvars varsc;
+    c
   in
+  let v = f nf_constr in
   let sigma = restrict_universe_context sigma !uvars in
   sigma, v
 
@@ -64,15 +67,7 @@ let kind_of_term_upto = EConstr.kind_upto
 let nf_evars_universes sigma t = EConstr.to_constr ~abort_on_undefined_evars:false sigma (EConstr.of_constr t)
 let whd_evar = EConstr.whd_evar
 
-let nf_evar sigma c =
-  let lsubst = Evd.universe_subst sigma in
-  let evar_value ev = Evd.existential_opt_value0 sigma ev in
-  let level_value l =
-    UnivSubst.level_subst_of (fun l -> UnivSubst.normalize_univ_variable_opt_subst lsubst l) l
-  in
-  let sort_value s = UState.nf_sort (Evd.evar_universe_context sigma) s in
-  let rel_value r = UState.nf_relevance (Evd.evar_universe_context sigma) r in
-  EConstr.of_constr @@ UnivSubst.nf_evars_and_universes_opt_subst evar_value level_value sort_value rel_value (EConstr.Unsafe.to_constr c)
+let nf_evar = Evd.MiniEConstr.nf_evar
 
 let j_nf_evar sigma j =
   { uj_val = nf_evar sigma j.uj_val;
@@ -86,10 +81,11 @@ let nf_relevance sigma r =
   UState.nf_relevance (Evd.evar_universe_context sigma) r
 
 let nf_named_context_evar sigma ctx =
-  Context.Named.map (nf_evars_universes sigma) ctx
+  Context.Named.map_with_relevance (nf_relevance sigma) (nf_evars_universes sigma) ctx
 
 let nf_rel_context_evar sigma ctx =
-  Context.Rel.map (nf_evar sigma) ctx
+  let nf_relevance r = ERelevance.make (ERelevance.kind sigma r) in
+  Context.Rel.map_with_relevance nf_relevance (nf_evar sigma) ctx
 
 let nf_env_evar sigma env =
   let nc' = nf_named_context_evar sigma (Environ.named_context env) in
@@ -141,7 +137,7 @@ let head_evar sigma c =
     | Case (_, _, _, _, _, c, _) -> hrec c
     | App (c,_)      -> hrec c
     | Cast (c,_,_)   -> hrec c
-    | Proj (p, c)    -> hrec c
+    | Proj (_, _, c)    -> hrec c
     | _              -> raise NoHeadEvar
   in
   hrec c
@@ -414,8 +410,7 @@ let next_evar_name sigma naming = match naming with
 | IntroAnonymous -> None
 | IntroIdentifier id -> Some id
 | IntroFresh id ->
-  let has_name id = try let _ = Evd.evar_key id sigma in true with Not_found -> false in
-  let id = Namegen.next_ident_away_from id has_name in
+  let id = Nameops.Fresh.next id (Evd.evar_names sigma) in
   Some id
 
 (* [new_evar] declares a new existential in an env env with type typ *)
@@ -436,7 +431,7 @@ let new_evar ?src ?filter ?relevance ?abstract_arguments ?candidates ?(naming =
     | Some filter -> Filter.filter_slist filter instance in
   let relevance = match relevance with
   | Some r -> r
-  | None -> Sorts.Relevant (* FIXME: relevant_of_type not defined yet *)
+  | None -> ERelevance.relevant (* FIXME: relevant_of_type not defined yet *)
   in
   let (evd, evk) = new_pure_evar sign evd typ' ?src ?filter ~relevance ?abstract_arguments ?candidates ?name
     ?typeclass_candidate ?principal in
@@ -444,7 +439,7 @@ let new_evar ?src ?filter ?relevance ?abstract_arguments ?candidates ?(naming =
 
 let new_type_evar ?src ?filter ?naming ?principal ?hypnaming env evd rigid =
   let (evd', s) = new_sort_variable rigid evd in
-  let relevance = EConstr.ESorts.relevance_of_sort evd s in
+  let relevance = EConstr.ESorts.relevance_of_sort s in
   let (evd', e) = new_evar env evd' ?src ?filter ~relevance ?naming ~typeclass_candidate:false ?principal ?hypnaming (EConstr.mkSort s) in
   evd', (e, s)
 
@@ -500,12 +495,12 @@ let filter_effective_candidates evd evi filter candidates =
   let ids = set_of_evctx (Filter.filter_list filter (evar_context evi)) in
   List.filter (fun a -> Id.Set.subset (collect_vars evd a) ids) candidates
 
-let restrict_evar evd evk filter ?src candidates =
+let restrict_evar evd evk filter candidates =
   let evar_info = Evd.find_undefined evd evk in
   let candidates = Option.map (filter_effective_candidates evd evar_info filter) candidates in
   match candidates with
   | Some [] -> raise (ClearDependencyError (*FIXME*)(Id.of_string "blah", (NoCandidatesLeft evk), None))
-  | _ -> Evd.restrict evk filter ?candidates ?src evd
+  | _ -> Evd.restrict evk filter ?candidates evd
 
 let rec check_and_clear_in_constr ~is_section_variable env evdref err ids ~global c =
   (* returns a new constr where all the evars have been 'cleaned'
@@ -640,10 +635,11 @@ let clear_hyps2_in_evi env sigma hyps t concl ids =
    goal ([advance] is used to figure if a side effect has modified the
    goal) it terminates quickly. *)
 let rec advance sigma evk =
-  let EvarInfo evi = Evd.find sigma evk in
-  match Evd.evar_body evi with
-  | Evar_empty -> Some evk
-  | Evar_defined v ->
+  match Evd.find_defined sigma evk with
+  | None -> Some evk
+  | Some evi ->
+    match Evd.evar_body evi with
+    | Evar_defined v ->
       match is_aliased_evar sigma evk with
       | Some evk -> advance sigma evk
       | None -> None
@@ -706,7 +702,7 @@ let cached_evar_of_hyp cache sigma decl accu = match cache with
   in
   let (decl', evs) = !r in
   let evs =
-    if NamedDecl.equal (==) decl decl' then snd !r
+    if NamedDecl.equal (==) (==) decl decl' then snd !r
     else
       let fold c acc =
         let evs = undefined_evars_of_term sigma c in
@@ -763,14 +759,20 @@ let compare_cumulative_instances cv_pb variances u u' sigma =
   let open UnivProblem in
   let cstrs = Univ.Constraints.empty in
   let soft = Set.empty in
+  let qs, us = UVars.Instance.to_array u
+  and qs', us' = UVars.Instance.to_array u' in
+  let qcstrs = enforce_eq_qualities qs qs' Set.empty in
+  match Evd.add_universe_constraints sigma qcstrs with
+  | exception UGraph.UniverseInconsistency p -> Inr p
+  | sigma ->
   let cstrs, soft = Array.fold_left3 (fun (cstrs, soft) v u u' ->
-      let open Univ.Variance in
+      let open UVars.Variance in
       match v with
       | Irrelevant -> cstrs, Set.add (UWeak (u,u')) soft
       | Covariant when cv_pb == Conversion.CUMUL ->
         Univ.Constraints.add (u,Univ.Le,u') cstrs, soft
       | Covariant | Invariant -> Univ.Constraints.add (u,Univ.Eq,u') cstrs, soft)
-      (cstrs,soft) variances (Univ.Instance.to_array u) (Univ.Instance.to_array u')
+      (cstrs,soft) variances us us'
   in
   match Evd.add_constraints sigma cstrs with
   | sigma ->
@@ -779,11 +781,17 @@ let compare_cumulative_instances cv_pb variances u u' sigma =
 
 let compare_constructor_instances evd u u' =
   let open UnivProblem in
-  let soft =
-    Array.fold_left2 (fun cs u u' -> Set.add (UWeak (u,u')) cs)
-      Set.empty (Univ.Instance.to_array u) (Univ.Instance.to_array u')
-  in
-  Evd.add_universe_constraints evd soft
+  let qs, us = UVars.Instance.to_array u
+  and qs', us' = UVars.Instance.to_array u' in
+  let qcstrs = enforce_eq_qualities qs qs' Set.empty in
+  match Evd.add_universe_constraints evd qcstrs with
+  | exception UGraph.UniverseInconsistency p -> Inr p
+  | evd ->
+    let soft =
+      Array.fold_left2 (fun cs u u' -> Set.add (UWeak (u,u')) cs)
+        Set.empty us us'
+    in
+    Inl (Evd.add_universe_constraints evd soft)
 
 (** [eq_constr_univs_test ~evd ~extended_evd t u] tests equality of
     [t] and [u] up to existential variable instantiation and
