@@ -9,7 +9,6 @@
 (************************************************************************)
 
 open Names
-open Mod_subst
 open Genarg
 
 module Store = Store.Make ()
@@ -17,7 +16,7 @@ module Store = Store.Make ()
 type intern_variable_status = {
   intern_ids : Id.Set.t;
   notation_variable_status :
-    (bool ref * Notation_term.subscopes option ref *
+    (bool ref * Notation_term.subscopes option ref * Notation_term.notation_var_binders option ref *
        Notation_term.notation_var_internalization_type)
       Id.Map.t
 }
@@ -50,7 +49,6 @@ type glob_constr_and_expr = Glob_term.glob_constr * Constrexpr.constr_expr optio
 type glob_constr_pattern_and_expr = Id.Set.t * glob_constr_and_expr * Pattern.constr_pattern
 
 type ('raw, 'glb) intern_fun = glob_sign -> 'raw -> glob_sign * 'glb
-type 'glb subst_fun = substitution -> 'glb -> 'glb
 type 'glb ntn_subst_fun = Id.Set.t -> Glob_term.glob_constr Id.Map.t -> 'glb -> 'glb
 
 module InternObj =
@@ -60,13 +58,6 @@ struct
   let default _ = None
 end
 
-module SubstObj =
-struct
-  type ('raw, 'glb, 'top) obj = 'glb subst_fun
-  let name = "subst"
-  let default _ = None
-end
-
 module NtnSubstObj =
 struct
   type ('raw, 'glb, 'top) obj = 'glb ntn_subst_fun
@@ -75,7 +66,6 @@ struct
 end
 
 module Intern = Register (InternObj)
-module Subst = Register (SubstObj)
 module NtnSubst = Register (NtnSubstObj)
 
 let intern = Intern.obj
@@ -85,21 +75,32 @@ let generic_intern ist (GenArg (Rawwit wit, v)) =
   let (ist, v) = intern wit ist v in
   (ist, in_gen (glbwit wit) v)
 
-(** Substitution functions *)
+type ('raw,'glb) intern_pat_fun = ?loc:Loc.t -> ('raw,'glb) intern_fun
 
-let substitute = Subst.obj
-let register_subst0 = Subst.register0
+module InternPatObj = struct
+  type ('raw, 'glb, 'top) obj = ('raw, 'glb) intern_pat_fun
+  let name = "intern_pat"
+  let default tag =
+    Some (fun ?loc ->
+        let name = Genarg.(ArgT.repr (get_arg_tag tag)) in
+        CErrors.user_err ?loc Pp.(str "This quotation is not supported in tactic patterns (" ++ str name ++ str ")"))
+end
 
-let generic_substitute subs (GenArg (Glbwit wit, v)) =
-  in_gen (glbwit wit) (substitute wit subs v)
+module InternPat = Register (InternPatObj)
 
-let () = Hook.set Detyping.subst_genarg_hook generic_substitute
+let intern_pat = InternPat.obj
+
+let register_intern_pat = InternPat.register0
+
+let generic_intern_pat ?loc ist (GenArg (Rawwit wit, v)) =
+  let (ist, v) = intern_pat wit ?loc ist v in
+  (ist, in_gen (glbwit wit) v)
 
 (** Notation substitution *)
 
 let substitute_notation = NtnSubst.obj
 let register_ntn_subst0 = NtnSubst.register0
 
-let generic_substitute_notation avoid env (GenArg (Glbwit wit, v)) =
-  let v = substitute_notation wit avoid env v in
-  in_gen (glbwit wit) v
+let generic_substitute_notation avoid env (GenArg (Glbwit wit, v) as orig) =
+  let v' = substitute_notation wit avoid env v in
+  if v' == v then orig else in_gen (glbwit wit) v'
