@@ -25,18 +25,46 @@
 #include "coq_fix_code.h"
 
 #ifdef THREADED_CODE
-char ** coq_instr_table;
-char * coq_instr_base;
+
+static char ** coq_instr_table;
+static char * coq_instr_base;
+#define VALINSTR(instr) ((opcode_t)(coq_instr_table[instr] - coq_instr_base))
+
+void coq_init_thread_code(void ** instr_table, void * instr_base)
+{
+  coq_instr_table = (char **) instr_table;
+  coq_instr_base = (char *) instr_base;
+}
+
+#else
+#define VALINSTR(instr) instr
 #endif /*  THREADED_CODE */
 
+int coq_is_instruction(opcode_t instr1, opcode_t instr2)
+{
+  return instr1 == VALINSTR(instr2);
+}
 
 void * coq_stat_alloc (asize_t sz)
 {
   void * result = malloc (sz);
-  if (result == NULL) raise_out_of_memory ();
+  if (result == NULL) caml_raise_out_of_memory ();
   return result;
 }
 
+static opcode_t accu_instr;
+code_t accumulate = &accu_instr;
+
+value coq_accumulate(value unit)
+{
+  CAMLparam1(unit);
+  CAMLlocal1(res);
+  accu_instr = VALINSTR(ACCUMULATE);
+  res = caml_alloc_small(1, Abstract_tag);
+  Code_val(res) = accumulate;
+  CAMLreturn(res);
+}
+
 value coq_makeaccu (value i) {
   CAMLparam1(i);
   CAMLlocal1(res);
@@ -70,13 +98,6 @@ value coq_pushpop (value i) {
   }
 }
 
-value coq_is_accumulate_code(value code){
-  code_t q = Code_val(code);
-  int res;
-  res = Is_instruction(q,ACCUMULATE);
-  return Val_bool(res);
-}
-
 #ifdef ARCH_BIG_ENDIAN
 #define Reverse_32(dst,src) {                                               \
   char * _p, * _q;                                                          \
@@ -114,8 +135,8 @@ value coq_tcode_of_code (value code) {
       uint32_t i, sizes, const_size, block_size;
       COPY32(q,p); p++;
       sizes=*q++;
-      const_size = sizes & 0xFFFFFF;
-      block_size = sizes >> 24;
+      const_size = sizes >> 8;
+      block_size = sizes & 0xFF;
       sizes = const_size + block_size;
       for(i=0; i<sizes; i++) { COPY32(q,p); p++; q++; };
     } else if (instr == CLOSUREREC || instr==CLOSURECOFIX) {
