@@ -125,6 +125,13 @@ if (sp - num_args < coq_stack_threshold) {                                     \
 #define Coq_alloc_small(result, wosize, tag) Alloc_small(result, wosize, tag)
 #endif
 
+/* Beware: Alloc_small can only invoked for allocations smaller than
+   Max_young_wosize, which is usually 256. For larger allocations,
+   caml_alloc_shr needs to be used instead, and the block needs to be filled
+   using caml_initialize. Note that these two functions never trigger a GC,
+   so calling Setup_for_gc is not needed.
+*/
+
 /* If there is an asynchronous exception, we reset the vm */
 #define Handle_potential_exception(res) \
   if (Is_exception_result(res)) {       \
@@ -265,6 +272,16 @@ extern void caml_process_pending_signals(void);
 extern double coq_next_up(double);
 extern double coq_next_down(double);
 
+value coq_subst_instance(value arg, value tosubst)
+{
+  static const value * closure_f = NULL;
+  if (closure_f == NULL) {
+     /* First time around, look up by name */
+    closure_f = caml_named_value("coq_subst_instance");
+  }
+  return caml_callback2_exn(*closure_f, arg, tosubst);
+}
+
 /* The interpreter itself */
 
 value coq_interprete
@@ -401,6 +418,18 @@ value coq_interprete
         Next;
       }
 
+      Instruct(PUSHACCMANY) {
+        int first = *pc++;
+        int size = *pc++;
+        int i;
+        print_instr("PUSHACCMANY");
+        sp -= size;
+        for (i = 1; i < size; ++i) sp[i - 1] = sp[first + i];
+        sp[size - 1] = accu;
+        accu = sp[first];
+        Next;
+      }
+
       Instruct(POP){
         print_instr("POP");
         sp += *pc++;
@@ -463,6 +492,20 @@ value coq_interprete
         accu = Field(coq_env, 2 + *pc++);
         Next;
       }
+
+      Instruct(PUSHENVACCMANY) {
+        int first = *pc++;
+        int size = *pc++;
+        int i;
+        print_instr("PUSHENVACCMANY");
+        first += 2;
+        sp -= size;
+        for (i = 1; i < size; ++i) sp[i - 1] = Field(coq_env, first + i);
+        sp[size - 1] = accu;
+        accu = Field(coq_env, first);
+        Next;
+      }
+
       /* Function application */
 
       Instruct(PUSH_RETADDR) {
@@ -765,8 +808,6 @@ value coq_interprete
               for (i = 3; i < sz; ++i)
                 Field(block, i) = *sp++;
             } else {
-              // too large for Alloc_small, so use caml_alloc_shr instead
-              // it never triggers a GC, so no need for Setup_for_gc
               block = caml_alloc_shr(sz, Closure_tag);
               caml_initialize(&Field(block, 2), accu);
               for (i = 3; i < sz; ++i)
@@ -819,7 +860,7 @@ value coq_interprete
         *--sp=accu;
         Coq_alloc_small(accu, nfuncs * 3 + nvars, Closure_tag);
         Field(accu, nfuncs * 3 + nvars - 1) = *sp++;
-        p = &Field(accu, 0);
+        p = (value *) &Field(accu, 0);
         *p++ = (value) (pc + pc[0]);
         *p++ = Val_int(nfuncs * 3 - 1);
         for (i = 1; i < nfuncs; i++) {
@@ -935,16 +976,21 @@ value coq_interprete
         print_instr("MAKEBLOCK, tag=");
         if (wosize == 0) {
           accu = Atom(tag);
-        } else {
+        } else if (wosize <= Max_young_wosize) {
           Coq_alloc_small(block, wosize, tag);
           Field(block, 0) = accu;
           for (i = 1; i < wosize; i++) Field(block, i) = *sp++;
           accu = block;
+        } else {
+          block = caml_alloc_shr(wosize, tag);
+          caml_initialize(&Field(block, 0), accu);
+          for (i = 1; i < wosize; i++) caml_initialize(&Field(block, i), *sp++);
+          accu = block;
         }
         Next;
       }
-      Instruct(MAKEBLOCK1) {
 
+      Instruct(MAKEBLOCK1) {
         tag_t tag = *pc++;
         value block;
         print_instr("MAKEBLOCK1, tag=");
@@ -1001,13 +1047,12 @@ value coq_interprete
       Instruct(SWITCH) {
         uint32_t sizes = *pc++;
         print_instr("SWITCH");
-        print_int(sizes & 0xFFFFFF);
         if (Is_block(accu)) {
           long index = Tag_val(accu);
           if (index == Closure_tag) index = 0;
           print_instr("block");
           print_lint(index);
-          pc += pc[(sizes & 0xFFFFFF) + index];
+          pc += pc[(sizes >> 8) + index];
         } else {
           long index = Long_val(accu);
           print_instr("constant");
@@ -1167,8 +1212,6 @@ value coq_interprete
           for (i = size; i < sz; ++i)
             Field(accu, i) = *sp++;
         } else {
-          // too large for Alloc_small, so use caml_alloc_shr instead
-          // it never triggers a GC, so no need for Setup_for_gc
           accu = caml_alloc_shr(sz, Closure_tag);
           for (i = 0; i < size; ++i)
             caml_initialize(&Field(accu, i), Field(coq_env, i));
@@ -1235,13 +1278,20 @@ value coq_interprete
             *--sp=Field(coq_global_data, annot);
             /* We save the stack */
             if (sz == 0) accu = Atom(0);
-            else {
+            else if (sz <= Max_young_wosize) {
               Coq_alloc_small(accu, sz, Default_tag);
               if (Is_tailrec_switch(*sp)) {
                 for (i = 0; i < sz; i++) Field(accu, i) = sp[i+2];
               }else{
                 for (i = 0; i < sz; i++) Field(accu, i) = sp[i+5];
               }
+            } else {
+              accu = caml_alloc_shr(sz, Default_tag);
+              if (Is_tailrec_switch(*sp)) {
+                for (i = 0; i < sz; i++) caml_initialize(&Field(accu, i), sp[i+2]);
+              }else{
+                for (i = 0; i < sz; i++) caml_initialize(&Field(accu, i), sp[i+5]);
+              }
             }
             *--sp = accu;
             /* Create bytecode wrappers */
@@ -1287,8 +1337,6 @@ value coq_interprete
           for (i = 3; i < sz; ++i)
             Field(accu, i) = *sp++;
         } else {
-          // too large for Alloc_small, so use caml_alloc_shr instead
-          // it never triggers a GC, so no need for Setup_for_gc
           accu = caml_alloc_shr(sz, Closure_tag);
           caml_initialize(&Field(accu, 2), Field(coq_atom_tbl, *pc));
           for (i = 3; i < sz; ++i)
@@ -1303,6 +1351,17 @@ value coq_interprete
         Next;
       }
 
+      Instruct(SUBSTINSTANCE) {
+        print_instr("SUBSTINSTANCE");
+        print_int(*pc);
+
+        Setup_for_caml_call;
+        accu = coq_subst_instance(accu, Field(coq_global_data, *pc++));
+        Restore_after_caml_call;
+        Handle_potential_exception(accu);
+        Next;
+      }
+
       Instruct(BRANCH) {
         /* unconditional branching */
         print_instr("BRANCH");
@@ -1930,6 +1989,26 @@ value coq_interprete
         Next;
       }
 
+      Instruct(CHECKCAMLCALL3) {
+        // arity-3 callback, no argument can be an accumulator
+        value arg1;
+        value arg2;
+        print_instr("CHECKCAMLCALL3");
+        if (!Is_accu(accu) && !Is_accu(sp[0]) && !Is_accu(sp[1])) {
+          pc++;
+          arg1 = sp[0];
+          arg2 = sp[1];
+          Setup_for_caml_call;
+          print_int(*pc);
+          accu = caml_callback3_exn(Field(coq_global_data, *pc), accu, arg1, arg2);
+          Restore_after_caml_call;
+          Handle_potential_exception(accu);
+          sp += 2;
+          pc++;
+        } else pc += *pc;
+        Next;
+      }
+
       Instruct(CHECKCAMLCALL3_1) {
         // arity-3 callback, the last argument can be an accumulator
         value arg1;
