@@ -8,6 +8,7 @@
 /*                                                                     */
 /***********************************************************************/
 
+#include <stdlib.h>
 #include <stdio.h>
 #include <caml/memory.h>
 #include "coq_fix_code.h"
@@ -20,6 +21,8 @@
 #define Setup_for_gc
 #define Restore_after_gc
 
+#define Is_instruction(c, i) coq_is_instruction(*c, i)
+
 value coq_kind_of_closure(value v) {
   opcode_t * c;
   int is_app = 0;
@@ -31,6 +34,13 @@ value coq_kind_of_closure(value v) {
   return Val_int(0);
 }
 
+value coq_is_accumulate_code(value code)
+{
+  code_t q = Code_val(code);
+  int res;
+  res = Is_instruction(q,ACCUMULATE);
+  return Val_bool(res);
+}
 
 /* DESTRUCT ACCU */
 
@@ -96,3 +106,14 @@ value coq_tcode_array(value tcodes) {
   }
   CAMLreturn(res);
 }
+
+CAMLprim value coq_obj_set_tag (value arg, value new_tag)
+{
+#if OCAML_VERSION >= 50000
+// Placeholder used by native_compute
+  abort();
+#else
+  Tag_val (arg) = Int_val (new_tag);
+#endif
+  return Val_unit;
+}
