@@ -56,15 +56,15 @@ let hcons_template_universe ar =
     template_context = Univ.hcons_universe_context_set ar.template_context }
 
 let universes_context = function
-  | Monomorphic -> Univ.AbstractContext.empty
+  | Monomorphic -> UVars.AbstractContext.empty
   | Polymorphic ctx -> ctx
 
 let abstract_universes = function
   | Entries.Monomorphic_entry ->
-    Univ.empty_level_subst, Monomorphic
+    UVars.empty_sort_subst, Monomorphic
   | Entries.Polymorphic_entry uctx ->
-    let (inst, auctx) = Univ.abstract_universes uctx in
-    let inst = Univ.make_instance_subst inst in
+    let (inst, auctx) = UVars.abstract_universes uctx in
+    let inst = UVars.make_instance_subst inst in
     (inst, Polymorphic auctx)
 
 (** {6 Constants } *)
@@ -76,7 +76,7 @@ let constant_is_polymorphic cb =
 
 
 let constant_has_body cb = match cb.const_body with
-  | Undef _ | Primitive _ -> false
+  | Undef _ | Primitive _ | Symbol _ -> false
   | Def _ | OpaqueDef _ -> true
 
 let constant_polymorphic_context cb =
@@ -84,7 +84,7 @@ let constant_polymorphic_context cb =
 
 let is_opaque cb = match cb.const_body with
   | OpaqueDef _ -> true
-  | Undef _ | Def _ | Primitive _ -> false
+  | Undef _ | Def _ | Primitive _ | Symbol _ -> false
 
 (** {7 Constant substitutions } *)
 
@@ -100,13 +100,13 @@ let subst_const_type subst arity =
 (** No need here to check for physical equality after substitution,
     at least for Def due to the delayed substitution [subst_constr_subst]. *)
 let subst_const_def subst def = match def with
-  | Undef _ | Primitive _ -> def
+  | Undef _ | Primitive _ | Symbol _ -> def
   | Def c -> Def (subst_mps subst c)
   | OpaqueDef o -> OpaqueDef (Opaqueproof.subst_opaque subst o)
 
 let subst_const_body subst cb =
   (* we're outside sections *)
-  assert (List.is_empty cb.const_hyps && Univ.Instance.is_empty cb.const_univ_hyps);
+  assert (List.is_empty cb.const_hyps && UVars.Instance.is_empty cb.const_univ_hyps);
   if is_empty_subst subst then cb
   else
     let body' = subst_const_def subst cb.const_body in
@@ -115,7 +115,7 @@ let subst_const_body subst cb =
     then cb
     else
       { const_hyps = [];
-        const_univ_hyps = Univ.Instance.empty;
+        const_univ_hyps = UVars.Instance.empty;
         const_body = body';
         const_type = type';
         const_body_code =
@@ -139,6 +139,7 @@ let hcons_rel_context l = List.Smart.map hcons_rel_decl l
 let hcons_const_def = function
   | Undef inl -> Undef inl
   | Primitive p -> Primitive p
+  | Symbol r -> Symbol r
   | Def l_constr ->
     Def (Constr.hcons l_constr)
   | OpaqueDef _ as x -> x (* hashconsed when turned indirect *)
@@ -147,7 +148,7 @@ let hcons_universes cbu =
   match cbu with
   | Monomorphic -> Monomorphic
   | Polymorphic ctx ->
-    Polymorphic (Univ.hcons_abstract_universe_context ctx)
+    Polymorphic (UVars.hcons_abstract_universe_context ctx)
 
 let hcons_const_body cb =
   { cb with
@@ -157,53 +158,49 @@ let hcons_const_body cb =
   }
 
 (** {6 Inductive types } *)
-let eq_nested_type t1 t2 = match t1, t2 with
-| NestedInd ind1, NestedInd ind2 -> Names.Ind.CanOrd.equal ind1 ind2
-| NestedInd _, _ -> false
-| NestedPrimitive c1, NestedPrimitive c2 -> Names.Constant.CanOrd.equal c1 c2
-| NestedPrimitive _, _ -> false
+
+let eq_recarg_type t1 t2 = match t1, t2 with
+| RecArgInd ind1, RecArgInd ind2 -> Names.Ind.CanOrd.equal ind1 ind2
+| RecArgPrim c1, RecArgPrim c2 -> Names.Constant.CanOrd.equal c1 c2
+| (RecArgInd _ | RecArgPrim _), _ -> false
 
 let eq_recarg r1 r2 = match r1, r2 with
 | Norec, Norec -> true
 | Norec, _ -> false
-| Mrec i1, Mrec i2 -> Names.Ind.CanOrd.equal i1 i2
+| Mrec t1, Mrec t2 -> eq_recarg_type t1 t2
 | Mrec _, _ -> false
-| Nested ty1, Nested ty2 -> eq_nested_type ty1 ty2
-| Nested _, _ -> false
 
-let pp_recarg = let open Pp in function
-  | Declarations.Norec -> str "Norec"
-  | Declarations.Mrec (mind,i) ->
+let pr_recarg_type = let open Pp in function
+  | RecArgInd (mind,i) ->
      str "Mrec[" ++ Names.MutInd.print mind ++ pr_comma () ++ int i ++ str "]"
-  | Declarations.(Nested (NestedInd (mind,i))) ->
-     str "Nested[" ++ Names.MutInd.print mind ++ pr_comma () ++ int i ++ str "]"
-  | Declarations.(Nested (NestedPrimitive c)) ->
-     str "Nested[" ++ Names.Constant.print c ++ str "]"
+  | RecArgPrim c ->
+     str "Prim[" ++ Names.Constant.print c ++ str "]"
 
-let pp_wf_paths x = Rtree.pp_tree pp_recarg x
+let pr_recarg = let open Pp in function
+  | Declarations.Norec -> str "Norec"
+  | Declarations.Mrec t -> pr_recarg_type t
 
-let subst_nested_type subst ty = match ty with
-| NestedInd (kn,i) ->
+let pr_wf_paths x = Rtree.pr_tree pr_recarg x
+
+let subst_recarg_type subst ty = match ty with
+| RecArgInd (kn,i) ->
   let kn' = subst_mind subst kn in
-  if kn==kn' then ty else NestedInd (kn',i)
-| NestedPrimitive c ->
+  if kn==kn' then ty else RecArgInd (kn',i)
+| RecArgPrim c ->
   let c',_ = subst_con subst c in
-  if c==c' then ty else NestedPrimitive c'
+  if c==c' then ty else RecArgPrim c'
 
 let subst_recarg subst r = match r with
   | Norec -> r
-  | Mrec (kn,i) ->
-    let kn' = subst_mind subst kn in
-    if kn==kn' then r else Mrec (kn',i)
-  | Nested ty ->
-    let ty' = subst_nested_type subst ty in
-    if ty==ty' then r else Nested ty'
+  | Mrec ty ->
+    let ty' = subst_recarg_type subst ty in
+    if ty==ty' then r else Mrec ty'
 
 let mk_norec = Rtree.mk_node Norec [||]
 
 let mk_paths r recargs =
   Rtree.mk_node r
-    (Array.map (fun l -> Rtree.mk_node Norec (Array.of_list l)) recargs)
+    (Array.map Array.of_list recargs)
 
 let dest_recarg p = fst (Rtree.dest_node p)
 
@@ -215,11 +212,11 @@ let dest_recarg p = fst (Rtree.dest_node p)
 let dest_subterms p =
   let (ra,cstrs) = Rtree.dest_node p in
   assert (match ra with Norec -> false | _ -> true);
-  Array.map (fun t -> Array.to_list (snd (Rtree.dest_node t))) cstrs
+  Array.map Array.to_list cstrs
 
 let recarg_length p j =
   let (_,cstrs) = Rtree.dest_node p in
-  Array.length (snd (Rtree.dest_node cstrs.(j-1)))
+  Array.length cstrs.(j-1)
 
 let subst_wf_paths subst p = Rtree.Smart.map (subst_recarg subst) p
 
@@ -247,7 +244,7 @@ let subst_mind_packet subst mbp =
     mind_user_lc = Array.Smart.map (subst_mps subst) mbp.mind_user_lc;
     mind_nrealargs = mbp.mind_nrealargs;
     mind_nrealdecls = mbp.mind_nrealdecls;
-    mind_kelim = mbp.mind_kelim;
+    mind_squashed = mbp.mind_squashed;
     mind_recargs = subst_wf_paths subst mbp.mind_recargs (*wf_paths*);
     mind_relevance = mbp.mind_relevance;
     mind_nb_constant = mbp.mind_nb_constant;
@@ -268,12 +265,12 @@ let subst_mind_record subst r = match r with
 
 let subst_mind_body subst mib =
   (* we're outside sections *)
-  assert (List.is_empty mib.mind_hyps && Univ.Instance.is_empty mib.mind_univ_hyps);
+  assert (List.is_empty mib.mind_hyps && UVars.Instance.is_empty mib.mind_univ_hyps);
   { mind_record = subst_mind_record subst mib.mind_record ;
     mind_finite = mib.mind_finite ;
     mind_ntypes = mib.mind_ntypes ;
     mind_hyps = [];
-    mind_univ_hyps = Univ.Instance.empty;
+    mind_univ_hyps = UVars.Instance.empty;
     mind_nparams = mib.mind_nparams;
     mind_nparams_rec = mib.mind_nparams_rec;
     mind_params_ctxt =
@@ -303,45 +300,27 @@ let inductive_make_projection ind mib ~proj_arg =
   | NotRecord | FakeRecord ->
     CErrors.anomaly Pp.(str "inductive_make_projection: not a primitive record.")
   | PrimRecord infos ->
-    let _, labs, relevances, _ = infos.(snd ind) in
-    let proj_relevant = match relevances.(proj_arg) with
-    | Sorts.Irrelevant -> false
-    | Sorts.Relevant -> true
-    | Sorts.RelevanceVar _ -> assert false
-    in
+    let _, labs, rs, _ = infos.(snd ind) in
     if proj_arg < 0 || Array.length labs <= proj_arg
     then CErrors.anomaly Pp.(str "inductive_make_projection: invalid proj_arg.");
-    Names.Projection.Repr.make ind
-      ~proj_relevant
+    let p = Names.Projection.Repr.make ind
       ~proj_npars:mib.mind_nparams
       ~proj_arg
       labs.(proj_arg)
+    in
+    p, rs.(proj_arg)
 
 let inductive_make_projections ind mib =
   match mib.mind_record with
   | NotRecord | FakeRecord -> None
   | PrimRecord infos ->
     let _, labs, relevances, _ = infos.(snd ind) in
-    let projs = Array.mapi (fun proj_arg lab ->
-        let proj_relevant = match relevances.(proj_arg) with
-        | Sorts.Irrelevant -> false
-        | Sorts.Relevant -> true
-        | Sorts.RelevanceVar _ -> assert false
-        in
-        Names.Projection.Repr.make ind ~proj_relevant ~proj_npars:mib.mind_nparams ~proj_arg lab)
-        labs
+    let projs = Array.map2_i (fun proj_arg lab r ->
+        Names.Projection.Repr.make ind ~proj_npars:mib.mind_nparams ~proj_arg lab, r)
+        labs relevances
     in
     Some projs
 
-let relevance_of_projection_repr mib p =
-  let _mind,i = Names.Projection.Repr.inductive p in
-  match mib.mind_record with
-  | NotRecord | FakeRecord ->
-    CErrors.anomaly ~label:"relevance_of_projection" Pp.(str "not a projection")
-  | PrimRecord infos ->
-    let _,_,rs,_ = infos.(i) in
-    rs.(Names.Projection.Repr.arg p)
-
 (** {6 Hash-consing of inductive declarations } *)
 
 let hcons_regular_ind_arity a =
@@ -374,6 +353,15 @@ let hcons_mind mib =
     mind_template = Option.Smart.map hcons_template_universe mib.mind_template;
     mind_universes = hcons_universes mib.mind_universes }
 
+let subst_rewrite_rules subst ({ rewrules_rules } as rules) =
+  let body' = List.Smart.map (fun (name, ({ rhs; _ } as rule) as orig) ->
+      let rhs' = subst_mps subst rhs in
+      if rhs == rhs' then orig else name, { rule with rhs = rhs' })
+      rewrules_rules
+  in
+  if rewrules_rules == body' then rules else
+    { rewrules_rules = body' }
+
 (** Hashconsing of modules *)
 
 let hcons_functorize hty he hself f = match f with
@@ -411,6 +399,7 @@ let rec hcons_structure_field_body sb = match sb with
 | SFBmodtype mb ->
   let mb' = hcons_module_type mb in
   if mb == mb' then sb else SFBmodtype mb'
+| SFBrules _ -> sb (* TODO? *)
 
 and hcons_structure_body sb =
   (** FIXME *)
