@@ -37,6 +37,7 @@ let opcodes =
     "PUSHACC6", 0;
     "PUSHACC7", 0;
     "PUSHACC", 1;
+    "PUSHACCMANY", 2;
     "POP", 1;
     "ENVACC0", 0;
     "ENVACC1", 0;
@@ -48,6 +49,7 @@ let opcodes =
     "PUSHENVACC2", 0;
     "PUSHENVACC3", 0;
     "PUSHENVACC", 1;
+    "PUSHENVACCMANY", 2;
     "PUSH_RETADDR", 1;
     "APPLY", 1;
     "APPLY1", 0;
@@ -99,6 +101,7 @@ let opcodes =
     "ACCUMULATE", 0;
     "MAKESWITCHBLOCK", 4;
     "MAKEACCU", 1;
+    "SUBSTINSTANCE", 1;
     "BRANCH", 1;
     "CHECKADDINT63", 1;
     "CHECKADDCINT63", 1;
@@ -154,15 +157,16 @@ let opcodes =
     "CHECKCAMLCALL2_1", 2;
     "CHECKCAMLCALL1", 2;
     "CHECKCAMLCALL2", 2;
+    "CHECKCAMLCALL3", 2;
     "CHECKCAMLCALL3_1", 2;
     "STOP", 0
   |]
 
 let pp_c_comment fmt =
-  Format.fprintf fmt "/* %s */"
+  Format.fprintf fmt "/* %s */\n"
 
 let pp_ocaml_comment fmt =
-  Format.fprintf fmt "(* %s *)"
+  Format.fprintf fmt "(* %s *)\n"
 
 let pp_header isOcaml fmt =
   Format.fprintf fmt "%a"
@@ -197,6 +201,12 @@ let pp_vmopcodes_ml fmt =
       Format.fprintf fmt "let op%s = %d@.@." s n
     ) (Array.map fst opcodes)
 
+let pp_vmopcodes_mli fmt =
+  pp_header true fmt;
+  Array.iteri (fun _ s ->
+      Format.fprintf fmt "val op%s : int@.@." s
+    ) (Array.map fst opcodes)
+
 let usage () =
   Format.eprintf "usage: %s [enum|jump|arity|copml]@." Sys.argv.(0);
   exit 1
@@ -207,6 +217,7 @@ let main () =
   | "jump" -> pp_coq_jumptbl_h Format.std_formatter
   | "arity" -> pp_coq_arity_h Format.std_formatter
   | "copml" -> pp_vmopcodes_ml Format.std_formatter
+  | "copmli" -> pp_vmopcodes_mli Format.std_formatter
   | _ -> usage ()
   | exception Invalid_argument _ -> usage ()
 
