@@ -264,7 +264,10 @@ sig
   val dp : t -> DirPath.t
 
   val to_string : t -> string
-  (** Encode as a string (not to be used for user-facing messages). *)
+  (** Converts a identifier into an string. *)
+
+  val print : t -> Pp.t
+  (** Pretty-printer. *)
 
   val debug_to_string : t -> string
   (** Same as [to_string], but outputs extra information related to debug. *)
@@ -306,9 +309,9 @@ sig
   val hash : t -> int
 end
 
-module KNset  : CSig.SetS with type elt = KerName.t
+module KNset  : CSig.USetS with type elt = KerName.t
 module KNpred : Predicate.S with type elt = KerName.t
-module KNmap  : Map.ExtS with type key = KerName.t and module Set := KNset
+module KNmap  : Map.UExtS with type key = KerName.t and module Set := KNset
 
 (** {6 Signature for quotiented names} *)
 
@@ -362,6 +365,9 @@ sig
 
   module SyntacticOrd : EqType with type t = t
   (** Equality functions using both names, for low-level uses. *)
+
+  val canonize : t -> t
+  (** Returns the canonical version of the name *)
 end
 
 (** {6 Constant Names } *)
@@ -379,7 +385,7 @@ sig
   (** Special case of [make] where the user name is canonical.  *)
 
   val make2 : ModPath.t -> Label.t -> t
-  (** Shortcut for [(make1 (KerName.make2 ...))] *)
+  (** Shortcut for [(make1 (KerName.make ...))] *)
 
   (** Projections *)
 
@@ -424,14 +430,14 @@ end
 (** The [*_env] modules consider an order on user part of names
    the others consider an order on canonical part of names*)
 module Cpred : Predicate.S with type elt = Constant.t
-module Cset : CSig.SetS with type elt = Constant.t
-module Cset_env  : CSig.SetS with type elt = Constant.t
+module Cset : CSig.USetS with type elt = Constant.t
+module Cset_env  : CSig.USetS with type elt = Constant.t
 
-module Cmap : Map.ExtS with type key = Constant.t and module Set := Cset
+module Cmap : Map.UExtS with type key = Constant.t and module Set := Cset
 (** A map whose keys are constants (values of the {!Constant.t} type).
     Keys are ordered wrt. "canonical form" of the constant. *)
 
-module Cmap_env : Map.ExtS with type key = Constant.t and module Set := Cset_env
+module Cmap_env : Map.UExtS with type key = Constant.t and module Set := Cset_env
 (** A map whose keys are constants (values of the {!Constant.t} type).
     Keys are ordered wrt. "user form" of the constant. *)
 
@@ -450,7 +456,7 @@ sig
   (** Special case of [make] where the user name is canonical.  *)
 
   val make2 : ModPath.t -> Label.t -> t
-  (** Shortcut for [(make1 (KerName.make2 ...))] *)
+  (** Shortcut for [(make1 (KerName.make ...))] *)
 
   (** Projections *)
 
@@ -488,9 +494,9 @@ sig
 
 end
 
-module Mindset : CSig.SetS with type elt = MutInd.t
-module Mindmap : Map.ExtS with type key = MutInd.t and module Set := Mindset
-module Mindmap_env : CMap.ExtS with type key = MutInd.t
+module Mindset : CSig.USetS with type elt = MutInd.t
+module Mindmap : Map.UExtS with type key = MutInd.t and module Set := Mindset
+module Mindmap_env : CMap.UExtS with type key = MutInd.t
 
 module Ind :
 sig
@@ -531,52 +537,10 @@ module Constrmap : CMap.ExtS with type key = constructor and module Set := Const
 module Indmap_env : CMap.ExtS with type key = inductive and module Set := Indset_env
 module Constrmap_env : CMap.ExtS with type key = constructor and module Set := Constrset_env
 
-val ind_modpath : inductive -> ModPath.t
-[@@ocaml.deprecated "Use the Ind module"]
-
-val constr_modpath : constructor -> ModPath.t
-[@@ocaml.deprecated "Use the Construct module"]
-
 val ith_mutual_inductive : inductive -> int -> inductive
 val ith_constructor_of_inductive : inductive -> int -> constructor
 val inductive_of_constructor : constructor -> inductive
 val index_of_constructor : constructor -> int
-val eq_ind : inductive -> inductive -> bool
-[@@ocaml.deprecated "Use the Ind module"]
-val eq_user_ind : inductive -> inductive -> bool
-[@@ocaml.deprecated "Use the Ind module"]
-val eq_syntactic_ind : inductive -> inductive -> bool
-[@@ocaml.deprecated "Use the Ind module"]
-val ind_ord : inductive -> inductive -> int
-[@@ocaml.deprecated "Use the Ind module"]
-val ind_hash : inductive -> int
-[@@ocaml.deprecated "Use the Ind module"]
-val ind_user_ord : inductive -> inductive -> int
-[@@ocaml.deprecated "Use the Ind module"]
-val ind_user_hash : inductive -> int
-[@@ocaml.deprecated "Use the Ind module"]
-val ind_syntactic_ord : inductive -> inductive -> int
-[@@ocaml.deprecated "Use the Ind module"]
-val ind_syntactic_hash : inductive -> int
-[@@ocaml.deprecated "Use the Ind module"]
-val eq_constructor : constructor -> constructor -> bool
-[@@ocaml.deprecated "Use the Construct module"]
-val eq_user_constructor : constructor -> constructor -> bool
-[@@ocaml.deprecated "Use the Construct module"]
-val eq_syntactic_constructor : constructor -> constructor -> bool
-[@@ocaml.deprecated "Use the Construct module"]
-val constructor_ord : constructor -> constructor -> int
-[@@ocaml.deprecated "Use the Construct module"]
-val constructor_hash : constructor -> int
-[@@ocaml.deprecated "Use the Construct module"]
-val constructor_user_ord : constructor -> constructor -> int
-[@@ocaml.deprecated "Use the Construct module"]
-val constructor_user_hash : constructor -> int
-[@@ocaml.deprecated "Use the Construct module"]
-val constructor_syntactic_ord : constructor -> constructor -> int
-[@@ocaml.deprecated "Use the Construct module"]
-val constructor_syntactic_hash : constructor -> int
-[@@ocaml.deprecated "Use the Construct module"]
 
 (** {6 Hash-consing } *)
 
@@ -597,6 +561,7 @@ type inv_rel_key = int (** index in the [rel_context] part of environment
                           of de Bruijn indice *)
 
 val eq_table_key : ('a -> 'a -> bool) -> 'a tableKey -> 'a tableKey -> bool
+val hash_table_key : ('a -> int) -> 'a tableKey -> int
 val eq_constant_key : Constant.t -> Constant.t -> bool
 
 (** equalities on constant and inductive names (for the checker) *)
@@ -615,7 +580,7 @@ module Projection : sig
   module Repr : sig
     type t
 
-    val make : inductive -> proj_npars:int -> proj_arg:int -> proj_relevant:bool -> Label.t -> t
+    val make : inductive -> proj_npars:int -> proj_arg:int -> Label.t -> t
 
     include QNameS with type t := t
 
@@ -627,7 +592,6 @@ module Projection : sig
     val npars : t -> int
     val arg : t -> int
     val label : t -> Label.t
-    val relevant : t -> bool
 
     val equal : t -> t -> bool [@@ocaml.deprecated "Use QProjection.equal"]
     val hash : t -> int [@@ocaml.deprecated "Use QProjection.hash"]
@@ -682,8 +646,20 @@ module Projection : sig
   val print : t -> Pp.t
   (** Print internal representation (not to be used for user-facing messages). *)
 
+  val debug_to_string : t -> string
+  (** Same as [to_string], but outputs extra information related to debug. *)
+
+  val debug_print : t -> Pp.t
+  (** Same as [print], but outputs extra information related to debug. *)
+
 end
 
+module PRset : CSig.USetS with type elt = Projection.Repr.t
+module PRmap : Map.UExtS with type key = Projection.Repr.t and module Set := PRset
+
+(** Predicate on projection representation (ignoring unfolding state) *)
+module PRpred : Predicate.S with type elt = Projection.Repr.t
+
 (** {6 Global reference is a kernel side type for all references together } *)
 
 module GlobRef : sig
@@ -701,12 +677,12 @@ module GlobRef : sig
 
   include QNameS with type t := t
 
-  module Set_env : CSig.SetS with type elt = t
-  module Map_env : Map.ExtS
+  module Set_env : CSig.USetS with type elt = t
+  module Map_env : Map.UExtS
     with type key = t and module Set := Set_env
 
-  module Set : CSig.SetS with type elt = t
-  module Map : Map.ExtS
+  module Set : CSig.USetS with type elt = t
+  module Map : Map.UExtS
     with type key = t and module Set := Set
 
   val print : t -> Pp.t
