@@ -8,13 +8,14 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 open Constr
-open Reduction
-open Nativelambda
+open Conversion
 
 (** This module implements the conversion test by compiling to OCaml code *)
 
-val native_conv : conv_pb -> evars -> types kernel_conversion_function
+val native_conv : conv_pb -> Genlambda.evars -> types kernel_conversion_function
 
 (** A conversion function parametrized by a universe comparator. Used outside of
     the kernel. *)
-val native_conv_gen : conv_pb -> evars -> (types, 'a) generic_conversion_function
+val native_conv_gen : conv_pb -> Genlambda.evars -> Environ.env -> ('a, 'err) generic_conversion_function
+
+val w_native_disabled : CWarnings.warning
