@@ -33,6 +33,11 @@ let add_retroknowledge retro action =
          | None -> { retro with retro_float64 = Some c }
          | Some c' -> check_same_types typ c c'; retro)
 
+      | PT_string ->
+        (match retro.retro_string with
+         | None -> { retro with retro_string = Some c }
+         | Some c' -> check_same_types typ c c'; retro)
+
       | PT_array ->
         (match retro.retro_array with
          | None -> { retro with retro_array = Some c }
@@ -95,6 +100,11 @@ let get_float_type env =
   | Some c -> c
   | None -> anomaly Pp.(str"Reduction of primitive: float64 not registered")
 
+let get_string_type env =
+  match env.retroknowledge.retro_string with
+  | Some c -> c
+  | None -> anomaly Pp.(str"Reduction of primitive: string not registered")
+
 let get_cmp_type env =
   match env.retroknowledge.retro_cmp with
   | Some (((mindcmp,_),_),_,_) ->
@@ -143,9 +153,11 @@ module type RedNativeEntries =
     val get : args -> int -> elem
     val get_int : evd -> elem -> Uint63.t
     val get_float : evd -> elem -> Float64.t
+    val get_string : evd -> elem -> Pstring.t
     val get_parray : evd -> elem -> elem Parray.t
     val mkInt : env -> Uint63.t -> elem
     val mkFloat : env -> Float64.t -> elem
+    val mkString : env -> Pstring.t -> elem
     val mkBool : env -> bool -> elem
     val mkCarry : env -> bool -> elem -> elem (* true if carry *)
     val mkIntPair : env -> elem -> elem -> elem
@@ -206,6 +218,8 @@ struct
 
   let get_parray evd args i = E.get_parray evd (E.get args i)
 
+  let get_string evd args i = E.get_string evd (E.get args i)
+
   let red_prim_aux env evd op u args =
     let open CPrimitives in
     match op with
@@ -390,6 +404,34 @@ struct
     | Arraylength ->
       let t = get_parray evd args 1 in
       E.mkInt env (Parray.length t)
+    | Stringmake ->
+      let i = get_int evd args 0 in
+      let c = get_int evd args 1 in
+      E.mkString env (Pstring.make i c)
+    | Stringlength ->
+      let s = get_string evd args 0 in
+      E.mkInt env (Pstring.length s)
+    | Stringget ->
+      let s = get_string evd args 0 in
+      let i = get_int evd args 1 in
+      E.mkInt env (Pstring.get s i)
+    | Stringsub ->
+      let s = get_string evd args 0 in
+      let off = get_int evd args 1 in
+      let len = get_int evd args 2 in
+      E.mkString env (Pstring.sub s off len)
+    | Stringcat ->
+      let s1 = get_string evd args 0 in
+      let s2 = get_string evd args 1 in
+      E.mkString env (Pstring.cat s1 s2)
+    | Stringcompare ->
+      let s1 = get_string evd args 0 in
+      let s2 = get_string evd args 1 in
+      begin match Pstring.compare s1 s2 with
+        | x when x < 0 -> E.mkLt env
+        | 0 -> E.mkEq env
+        | _ -> E.mkGt env
+      end
 
   let red_prim env evd p u args =
     try
