@@ -29,6 +29,12 @@ type caml_prim =
 | CAML_Arrayset
 | CAML_Arraycopy
 | CAML_Arraylength
+| CAML_Stringmake
+| CAML_Stringlength
+| CAML_Stringget
+| CAML_Stringsub
+| CAML_Stringcat
+| CAML_Stringcompare
 
 type instruction =
   | Klabel of Label.t
@@ -52,6 +58,7 @@ type instruction =
   | Kclosurecofix of int * int * Label.t array * Label.t array
                    (** nb fv, init, lbl types, lbl bodies *)
   | Kgetglobal of Constant.t
+  | Ksubstinstance of UVars.Instance.t
   | Kconst of structured_constant
   | Kmakeblock of (* size: *) int * tag (** allocate an ocaml block. Index 0
                                          ** is accu, all others are popped from
@@ -77,8 +84,6 @@ val pp_bytecodes : bytecodes -> Pp.t
 type fv_elem =
   FVnamed of Id.t
 | FVrel of int
-| FVuniv_var of int
-| FVevar of Evar.t
 
 type fv = fv_elem array
 
