@@ -103,32 +103,31 @@ type t = fv_elem
 
 let compare e1 e2 = match e1, e2 with
 | FVnamed id1, FVnamed id2 -> Id.compare id1 id2
-| FVnamed _, (FVrel _ | FVuniv_var _ | FVevar _) -> -1
+| FVnamed _, (FVrel _) -> -1
 | FVrel _, FVnamed _ -> 1
 | FVrel r1, FVrel r2 -> Int.compare r1 r2
-| FVrel _, (FVuniv_var _ | FVevar _) -> -1
-| FVuniv_var i1, FVuniv_var i2 -> Int.compare i1 i2
-| FVuniv_var _, (FVnamed _ | FVrel _) -> 1
-| FVuniv_var _, FVevar _ -> -1
-| FVevar _, (FVnamed _ | FVrel _ | FVuniv_var _) -> 1
-| FVevar e1, FVevar e2 -> Evar.compare e1 e2
 
 end
 
 module FvMap = Map.Make(Fv_elem)
 
+type fv_or_univ =
+| FUniv
+| FV of fv_elem
+
 (*spiwack: both type have been moved from Vmbytegen because I needed then
   for the retroknowledge *)
 type vm_env = {
-    size : int;              (* longueur de la liste [n] *)
-    fv_rev : fv_elem list;   (* [fvn; ... ;fv1] *)
-    fv_fwd : int FvMap.t;    (* reverse mapping *)
+    size : int;               (* longueur de la liste [n] *)
+    fv_rev : fv_or_univ list; (* [fvn; ... ;fv1] *)
+    fv_fwd : int FvMap.t;     (* reverse mapping *)
+    fv_unv : int option;      (* position of the universe instance *)
   }
 
 
 type comp_env = {
     arity : int;                 (* arity of the current function, 0 if none *)
-    nb_uni_stack : int ;         (* number of universes on the stack,      *)
+    toplevel_univs : bool;       (* is the toplevel instance on the stack  *)
                                  (* universes are always at the bottom.    *)
     nb_stack : int;              (* number of variables on the stack       *)
     in_stack : int Range.t;      (* position in the stack                  *)
@@ -142,6 +141,7 @@ type comp_env = {
 
 type glob_env = {
   env : Environ.env;
+  uinst_len : int * int; (** Size of the toplevel universe instance *)
   mutable fun_code : instruction list; (** Code of closures *)
 }
 
@@ -153,20 +153,21 @@ module Config = struct
   let stack_safety_margin = 15
 end
 
-type argument = ArgLambda of lambda | ArgUniv of Univ.Level.t
+type argument = ArgLambda of lambda | ArgInstance of UVars.Instance.t
 
-let empty_fv = { size= 0;  fv_rev = []; fv_fwd = FvMap.empty }
+let empty_fv = { size= 0;  fv_rev = []; fv_fwd = FvMap.empty; fv_unv = None }
 let push_fv d e = {
   size = e.size + 1;
-  fv_rev = d :: e.fv_rev;
+  fv_rev = (FV d) :: e.fv_rev;
   fv_fwd = FvMap.add d e.size e.fv_fwd;
+  fv_unv = e.fv_unv;
 }
 
 let fv r = !(r.in_env)
 
 let empty_comp_env ()=
   { arity = 0;
-    nb_uni_stack = 0;
+    toplevel_univs = false;
     nb_stack = 0;
     in_stack = Range.empty;
     pos_rec = [||];
@@ -192,9 +193,9 @@ let ensure_stack_capacity (cenv : comp_env) code =
 let rec add_param n sz l =
   if Int.equal n 0 then l else add_param (n - 1) sz (Range.cons (n+sz) l)
 
-let comp_env_fun ?(univs=0) arity =
+let comp_env_fun ?(univs=false) arity =
   { arity;
-    nb_uni_stack = univs ;
+    toplevel_univs = univs;
     nb_stack = arity;
     in_stack = add_param arity 0 Range.empty;
     pos_rec = [||];
@@ -206,7 +207,7 @@ let comp_env_fun ?(univs=0) arity =
 
 let comp_env_fix_type  rfv =
   { arity = 0;
-    nb_uni_stack = 0;
+    toplevel_univs = false;
     nb_stack = 0;
     in_stack = Range.empty;
     pos_rec = [||];
@@ -217,7 +218,7 @@ let comp_env_fix_type  rfv =
 
 let comp_env_fix ndef arity rfv =
    { arity;
-     nb_uni_stack = 0;
+     toplevel_univs = false;
      nb_stack = arity;
      in_stack = add_param arity 0 Range.empty;
      pos_rec = Array.init ndef (fun i -> Koffsetclosure i);
@@ -228,7 +229,7 @@ let comp_env_fix ndef arity rfv =
 
 let comp_env_cofix_type ndef rfv =
   { arity = 0;
-    nb_uni_stack = 0;
+    toplevel_univs = false;
     nb_stack = 0;
     in_stack = Range.empty;
     pos_rec = [||];
@@ -239,7 +240,7 @@ let comp_env_cofix_type ndef rfv =
 
 let comp_env_cofix ndef arity rfv =
    { arity;
-     nb_uni_stack = 0;
+     toplevel_univs = false;
      nb_stack = arity;
      in_stack = add_param arity 0 Range.empty;
      pos_rec = Array.init ndef (fun i -> Kenvacc (ndef - 1 - i));
@@ -290,34 +291,45 @@ let pos_rel i r sz =
         r.in_env := push_fv db env;
         Kenvacc(r.offset + pos)
 
-let pos_universe_var i r sz =
+let pos_instance r sz =
   (* Compilation of a universe variable can happen either at toplevel (the
   current closure correspond to a constant and has local universes) or in a
   local closure (which has no local universes). *)
-  if r.nb_uni_stack != 0 then
+  if r.toplevel_univs then
     (* Universe variables are represented by De Bruijn levels (not indices),
-    starting at 0. The shape of the stack will be [v1|..|vn|u1..up|arg1..argq]
-    with size = n + p + q, and q = r.arity. So Kacc (sz - r.arity - 1) will access
-    the last universe. *)
-    Kacc (sz - r.arity - (r.nb_uni_stack - i))
+    starting at 0. The shape of the stack will be [v1|..|vn|inst|arg1..argp]
+    with size = n + p + 1, and p = r.arity. So Kacc (sz - r.arity - 1) will access
+    the instance. *)
+    Kacc (sz - r.arity - 1)
   else
     let env = !(r.in_env) in
-    let db = FVuniv_var i in
-    try Kenvacc (r.offset + find_at db env)
-    with Not_found ->
+    let pos = match env.fv_unv with
+    | None ->
       let pos = env.size in
-      r.in_env := push_fv db env;
-      Kenvacc(r.offset + pos)
-
-let pos_evar evk r =
-  let env = !(r.in_env) in
-  let cid = FVevar evk in
-  try Kenvacc(r.offset + find_at cid env)
-  with Not_found ->
-    let pos = env.size in
-    r.in_env := push_fv cid env;
+      let env = {
+        size = pos + 1;
+        fv_rev = FUniv :: env.fv_rev;
+        fv_fwd = env.fv_fwd;
+        fv_unv = Some pos;
+      }
+      in
+      let () = r.in_env := env in
+      pos
+    | Some p -> p
+    in
     Kenvacc (r.offset + pos)
 
+let is_toplevel_inst env u =
+  UVars.eq_sizes env.uinst_len (UVars.Instance.length u)
+  && let qs, us = UVars.Instance.to_array u in
+  Array.for_all_i (fun i q -> Sorts.Quality.equal q (Sorts.Quality.var i)) 0 qs
+  && Array.for_all_i (fun i l -> Univ.Level.equal l (Univ.Level.var i)) 0 us
+
+let is_closed_inst u =
+  let qs, us = UVars.Instance.to_array u in
+  Array.for_all (fun q -> Option.is_empty (Sorts.Quality.var_index q)) qs
+  && Array.for_all (fun l -> Option.is_empty (Univ.Level.var_index l)) us
+
 (*i  Examination of the continuation *)
 
 (* Discard all instructions up to the next label.                        *)
@@ -476,10 +488,9 @@ let comp_app comp_fun comp_arg cenv f args sz cont =
 
 let compile_fv_elem cenv fv sz cont =
   match fv with
-  | FVrel i -> pos_rel i cenv sz :: cont
-  | FVnamed id -> pos_named id cenv :: cont
-  | FVuniv_var i -> pos_universe_var i cenv sz :: cont
-  | FVevar evk -> pos_evar evk cenv :: cont
+  | FV (FVrel i) -> pos_rel i cenv sz :: cont
+  | FV (FVnamed id) -> pos_named id cenv :: cont
+  | FUniv -> pos_instance cenv sz :: cont
 
 let rec compile_fv cenv l sz cont =
   match l with
@@ -513,6 +524,12 @@ let get_caml_prim = let open CPrimitives in function
   | Arrayset -> Some CAML_Arrayset
   | Arraycopy -> Some CAML_Arraycopy
   | Arraylength -> Some CAML_Arraylength
+  | Stringmake -> Some CAML_Stringmake
+  | Stringlength -> Some CAML_Stringlength
+  | Stringget -> Some CAML_Stringget
+  | Stringsub -> Some CAML_Stringsub
+  | Stringcat -> Some CAML_Stringcat
+  | Stringcompare -> Some CAML_Stringcompare
   | _ -> None
 
 (* sz is the size of the local stack *)
@@ -523,12 +540,14 @@ let rec compile_lam env cenv lam sz cont =
 
   | Lint i -> compile_structured_constant cenv (Const_b0 i) sz cont
 
-  | Lval v -> compile_structured_constant cenv (Const_val v) sz cont
+  | Lval v -> compile_structured_constant cenv (Const_val (get_lval v)) sz cont
 
   | Luint i -> compile_structured_constant cenv (Const_uint i) sz cont
 
   | Lfloat f -> compile_structured_constant cenv (Const_float f) sz cont
 
+  | Lstring s -> compile_structured_constant cenv (Const_string s) sz cont
+
   | Lproj (p,arg) ->
      compile_lam env cenv arg sz (Kproj (Projection.Repr.arg p) :: cont)
 
@@ -536,43 +555,42 @@ let rec compile_lam env cenv lam sz cont =
 
   | Levar (evk, args) ->
       if Array.is_empty args then
-        compile_fv_elem cenv (FVevar evk) sz cont
+        compile_structured_constant cenv (Const_evar evk) sz cont
       else
         (** Arguments are reversed in evar instances *)
         let args = Array.copy args in
         let () = Array.rev args in
-        comp_app compile_fv_elem (compile_lam env) cenv (FVevar evk) args sz cont
+        comp_app compile_structured_constant (compile_lam env) cenv (Const_evar evk) args sz cont
 
   | Lconst (kn,u) -> compile_constant env cenv kn u [||] sz cont
 
   | Lind (ind,u) ->
-    if Univ.Instance.is_empty u then
+    if UVars.Instance.is_empty u then
       compile_structured_constant cenv (Const_ind ind) sz cont
-    else comp_app compile_structured_constant compile_universe cenv
-        (Const_ind ind) (Univ.Instance.to_array u) sz cont
+    else comp_app compile_structured_constant (compile_instance env) cenv
+        (Const_ind ind) [|u|] sz cont
 
   | Lsort s ->
     (* We represent universes as a global constant with local universes
-       "compacted", i.e. as [u arg0 ... argn] where we will substitute (after
+       passed as the local universe instance, where we will substitute (after
        evaluation) [Var 0,...,Var n] with values of [arg0,...,argn] *)
-    let s, subs = match s with
-    | Sorts.Set | Sorts.Prop | Sorts.SProp as s -> s, []
+    let has_var = match s with
+    | Sorts.Set | Sorts.Prop | Sorts.SProp -> false
     | Sorts.Type u ->
-      let u, s = Univ.compact_univ u in
-      Sorts.sort_of_univ u, s
+      Univ.Universe.exists (fun (l, _) -> Option.has_some (Univ.Level.var_index l)) u
     | Sorts.QSort (q, u) ->
-      let u, s = Univ.compact_univ u in
-      Sorts.qsort q u, s
+      Option.has_some (Sorts.QVar.var_index q)
+      || Univ.Universe.exists (fun (l, _) -> Option.has_some (Univ.Level.var_index l)) u
     in
-    let compile_get_univ cenv idx sz cont =
+    let compile_instance cenv () sz cont =
       let () = set_max_stack_size cenv sz in
-      compile_fv_elem cenv (FVuniv_var idx) sz cont
+      pos_instance cenv sz :: cont
     in
-    if List.is_empty subs then
+    if not has_var then
       compile_structured_constant cenv (Const_sort s) sz cont
     else
-      comp_app compile_structured_constant compile_get_univ cenv
-        (Const_sort s) (Array.of_list subs) sz cont
+      comp_app compile_structured_constant compile_instance cenv
+        (Const_sort s) [|()|] sz cont
 
   | Llet (_id,def,body) ->
       compile_lam env cenv def sz
@@ -768,7 +786,17 @@ let rec compile_lam env cenv lam sz cont =
     (* Hack: brutally pierce through the abstraction of PArray *)
     let dummy = KerName.make (ModPath.MPfile DirPath.empty) (Names.Label.of_id @@ Id.of_string "dummy") in
     let dummy = (MutInd.make1 dummy, 0) in
-    let ar = Lmakeblock (dummy, 0, args) in (* build the ocaml array *)
+    let ar, cont = match Vmlambda.as_value 0 args with
+    | None ->
+      (* build the ocaml array *)
+      Lmakeblock (dummy, 0, args), cont
+    | Some v ->
+      (* dump the blob as is, but copy the resulting parray afterwards so that
+         the blob is left untouched by modifications to the resulting parray *)
+      let lbl = Label.create () in
+      (* dummy label, the array will never be an accumulator *)
+      Lval v, Klabel lbl :: Kcamlprim (CAML_Arraycopy, lbl) :: cont
+    in
     let kind = Lmakeblock (dummy, 0, [|ar; def|]) in (* Parray.Array *)
     let v = Lmakeblock (dummy, 0, [|kind|]) (* the reference *) in
     compile_lam env cenv v sz cont
@@ -785,7 +813,7 @@ let rec compile_lam env cenv lam sz cont =
       let lbl_default = Label.create () in
       let default =
         let cont = [Kshort_apply arity; jump] in
-        let cont = Kpush :: compile_get_global cenv kn (sz + arity) cont in
+        let cont = Kpush :: compile_get_global env cenv kn (sz + arity) cont in
         let cont =
           if Int.equal nparams 0 then cont
           else
@@ -799,55 +827,58 @@ let rec compile_lam env cenv lam sz cont =
       comp_args (compile_lam env) cenv args sz (Kprim(op, kn)::cont)
     end
 
-  | Lforce -> CErrors.anomaly Pp.(str "The VM should not use force")
-
-and compile_get_global cenv (kn,u) sz cont =
+and compile_get_global env cenv (kn,u) sz cont =
   let () = set_max_stack_size cenv sz in
-  if Univ.Instance.is_empty u then
+  if UVars.Instance.is_empty u then
     Kgetglobal kn :: cont
   else
     comp_app (fun _ _ _ cont -> Kgetglobal kn :: cont)
-      compile_universe cenv () (Univ.Instance.to_array u) sz cont
+      (compile_instance env) cenv () [|u|] sz cont
 
-and compile_universe cenv uni sz cont =
+and compile_instance env cenv u sz cont =
   let () = set_max_stack_size cenv sz in
-  match Univ.Level.var_index uni with
-  | None -> compile_structured_constant cenv (Const_univ_level uni) sz cont
-  | Some idx -> pos_universe_var idx cenv sz :: cont
+  if is_toplevel_inst env u then
+    (* Optimization: do not reallocate the same instance *)
+    pos_instance cenv sz :: cont
+  else if is_closed_inst u then
+    (* Optimization: allocate closed instances globally *)
+    compile_structured_constant cenv (Const_univ_instance u) sz cont
+  else
+    pos_instance cenv sz :: Ksubstinstance u :: cont
 
 and compile_constant env cenv kn u args sz cont =
   let () = set_max_stack_size cenv sz in
-  if Univ.Instance.is_empty u then
+  if UVars.Instance.is_empty u then
     (* normal compilation *)
     comp_app (fun _ _ sz cont ->
-        compile_get_global cenv (kn,u) sz cont)
+        compile_get_global env cenv (kn,u) sz cont)
       (compile_lam env) cenv () args sz cont
   else
     let compile_arg cenv constr_or_uni sz cont =
       match constr_or_uni with
       | ArgLambda t -> compile_lam env cenv t sz cont
-      | ArgUniv uni -> compile_universe cenv uni sz cont
+      | ArgInstance u -> compile_instance env cenv u sz cont
     in
-    let u = Univ.Instance.to_array u in
-    let lu = Array.length u in
     let all =
-      Array.init (lu + Array.length args)
-        (fun i -> if i < lu then ArgUniv u.(i) else ArgLambda args.(i-lu))
+      Array.init (Array.length args + 1)
+        (fun i -> if Int.equal i 0 then ArgInstance u else ArgLambda args.(i - 1))
     in
     comp_app (fun _ _ _ cont -> Kgetglobal kn :: cont)
       compile_arg cenv () all sz cont
 
-let is_univ_copy max u =
-  let u = Univ.Instance.to_array u in
-  if Array.length u = max then
-    Array.fold_left_i (fun i acc u ->
-        if acc then
-          match Univ.Level.var_index u with
-          | None -> false
-          | Some l -> l = i
-        else false) true u
-  else
-    false
+let coq_subst_instance : UVars.Instance.t -> UVars.Instance.t -> UVars.Instance.t =
+  UVars.subst_instance_instance
+
+let () = Callback.register "coq_subst_instance" coq_subst_instance
+
+let is_univ_copy (maxq,maxu) u =
+  let qs,us = UVars.Instance.to_array u in
+  let check_array max var_index a =
+    Array.length a = max
+    && Array.for_all_i (fun i x -> Option.equal Int.equal (var_index x) (Some i)) 0 a
+  in
+  check_array maxq Sorts.Quality.var_index qs
+  && check_array maxu Univ.Level.var_index us
 
 let dump_bytecode = ref false
 
@@ -860,15 +891,32 @@ let dump_bytecodes init code fvs =
      prlist_with_sep (fun () -> str "; ") pp_fv_elem fvs ++
      fnl ())
 
-let compile ~fail_on_error ?universes:(universes=0) env sigma c =
+let skip_suffix l =
+  let rec aux = function
+  | [] -> None
+  | b :: l ->
+    match aux l with
+    | None -> if b then None else Some [b]
+    | Some l -> Some (b :: l)
+  in
+  match aux l with None -> [] | Some l -> l
+
+let compile ?universes:(universes=(0,0)) env sigma c =
   Label.reset_label_counter ();
+  let lam = lambda_of_constr env sigma c in
+  let params, body = decompose_Llam lam in
+  let arity = Array.length params in
+  let mask =
+    let rels = Genlambda.free_rels body in
+    let init i = Int.Set.mem (arity - i) rels in
+    let mask = List.init arity init in
+    Array.of_list @@ skip_suffix mask
+  in
   let cont = [Kstop] in
-  try
     let cenv, init_code, fun_code =
-      if Int.equal universes 0 then
-        let lam = lambda_of_constr ~optimize:true env sigma c in
+      if UVars.eq_sizes universes (0,0) then
         let cenv = empty_comp_env () in
-        let env = { env; fun_code = [] } in
+        let env = { env; fun_code = []; uinst_len = (0,0) } in
         let cont = compile_lam env cenv lam 0 cont in
         let cont = ensure_stack_capacity cenv cont in
         cenv, cont, env.fun_code
@@ -876,14 +924,11 @@ let compile ~fail_on_error ?universes:(universes=0) env sigma c =
         (* We are going to generate a lambda, but merge the universe closure
          * with the function closure if it exists.
          *)
-        let lam = lambda_of_constr ~optimize:true env sigma c in
-        let params, body = decompose_Llam lam in
-        let arity = Array.length params in
         let cenv = empty_comp_env () in
-        let full_arity = arity + universes in
-        let r_fun = comp_env_fun ~univs:universes arity in
+        let full_arity = arity + 1 in
+        let r_fun = comp_env_fun ~univs:true arity in
         let lbl_fun = Label.create () in
-        let env = { env; fun_code = [] } in
+        let env = { env; fun_code = []; uinst_len = universes } in
         let cont_fun = compile_lam env r_fun body full_arity [Kreturn full_arity] in
         let cont_fun = ensure_stack_capacity r_fun cont_fun in
         let () = push_fun env (add_grab full_arity lbl_fun cont_fun) in
@@ -892,24 +937,37 @@ let compile ~fail_on_error ?universes:(universes=0) env sigma c =
         let init_code = ensure_stack_capacity cenv init_code in
         cenv, init_code, env.fun_code
     in
-    let fv = List.rev (!(cenv.in_env).fv_rev) in
+    let map_fv = function
+    | FV fv -> fv
+    | FUniv -> assert false
+    in
+    let fv = List.rev_map map_fv (!(cenv.in_env).fv_rev) in
     (if !dump_bytecode then
       Feedback.msg_debug (dump_bytecodes init_code fun_code fv)) ;
     let res = init_code @ fun_code in
-    Some (to_memory res, Array.of_list fv)
-  with TooLargeInductive msg as exn ->
-    let _, info = Exninfo.capture exn in
-    let fn = if fail_on_error then
-        CErrors.user_err ?loc:None ~info
-      else
-        (fun x -> Feedback.msg_warning x) in
-    fn msg; None
+    let code, patch = to_memory (Array.of_list fv) res in
+    mask, code, patch
+
+let warn_compile_error =
+  CWarnings.create ~name:"bytecode-compiler-failed-compilation" ~category:CWarnings.CoreCategories.bytecode_compiler
+    Vmerrors.pr_error
+
+let compile ~fail_on_error ?universes env sigma c =
+  try Some (compile ?universes env sigma c)
+  with Vmerrors.CompileError msg as exn ->
+    let exn = Exninfo.capture exn in
+    if fail_on_error then
+      Exninfo.iraise exn
+    else begin
+      warn_compile_error msg;
+      None
+    end
 
 let compile_constant_body ~fail_on_error env univs = function
   | Undef _ | OpaqueDef _ -> Some BCconstant
-  | Primitive _ -> None
+  | Primitive _ | Symbol _ -> None
   | Def body ->
-      let instance_size = Univ.AbstractContext.size (Declareops.universes_context univs) in
+      let instance_size = UVars.AbstractContext.size (Declareops.universes_context univs) in
       let alias =
         match kind body with
         | Const (kn',u) when is_univ_copy instance_size u ->
@@ -925,8 +983,8 @@ let compile_constant_body ~fail_on_error env univs = function
       match alias with
       | Some kn -> Some (BCalias kn)
       | _ ->
-          let res = compile ~fail_on_error ~universes:instance_size env empty_evars body in
-          Option.map (fun (code, fv) -> BCdefined (code, fv)) res
+        let res = compile ~fail_on_error ~universes:instance_size env (empty_evars env) body in
+        Option.map (fun (mask, code, patch) -> BCdefined (mask, code, patch)) res
 
 (* Shortcut of the previous function used during module strengthening *)
 
