@@ -11,11 +11,13 @@
 open Vmvalues
 open Environ
 
-type lambda = structured_values Genlambda.lambda
+type lval
+type lambda = lval Genlambda.lambda
 
-exception TooLargeInductive of Pp.t
+val get_lval : lval -> structured_values
+val as_value : int -> lambda array -> lval option
 
-val lambda_of_constr : optimize:bool -> env -> Genlambda.evars -> Constr.t -> lambda
+val lambda_of_constr : env -> Genlambda.evars -> Constr.t -> lambda
 
 (** Dump the VM lambda code after compilation (for debugging purposes) *)
 val dump_lambda : bool ref
