@@ -55,12 +55,14 @@ type 'a table_of_A =  {
   print : unit -> unit;
 }
 
+let opts_cat = Libobject.create_category "options"
+
 module MakeTable =
   functor
    (A : sig
           type t
           type key
-          module Set : CSig.SetS with type elt = t
+          module Set : CSig.USetS with type elt = t
           val table : (string * key table_of_A) list ref
           val encode : Environ.env -> key -> t
           val subst : Mod_subst.substitution -> t -> t
@@ -100,7 +102,7 @@ module MakeTable =
           Libobject.declare_object {(Libobject.default_object nick) with
                 Libobject.object_stage = Summary.Stage.Synterp;
                 Libobject.load_function = load_options;
-                Libobject.open_function = Libobject.simple_open load_options;
+                Libobject.open_function = Libobject.simple_open ~cat:opts_cat load_options;
                 Libobject.cache_function = cache_options;
                 Libobject.subst_function = subst_options;
                 Libobject.classify_function = (fun x -> Substitute)}
@@ -170,7 +172,7 @@ let get_ref_table k = String.List.assoc (nickname k) !ref_table
 module type RefConvertArg =
 sig
   type t
-  module Set : CSig.SetS with type elt = t
+  module Set : CSig.USetS with type elt = t
   val encode : Environ.env -> Libnames.qualid -> t
   val subst : Mod_subst.substitution -> t -> t
   val printer : t -> Pp.t
@@ -295,7 +297,7 @@ let declare_option cast uncast append ?(preprocess = fun x -> x)
           { (default_object (nickname key)) with
             object_stage = stage;
             load_function = load_options;
-            open_function = simple_open open_options;
+            open_function = simple_open ~cat:opts_cat open_options;
             cache_function = cache_options;
             subst_function = subst_options;
             discharge_function = discharge_options;
