@@ -87,6 +87,7 @@ module type FrozenStage = sig
   val make_marshallable : frozen -> frozen
   val unfreeze_summaries : ?partial:bool -> frozen -> unit
   val init_summaries : unit -> unit
+  val project_from_summary : frozen -> 'a Dyn.tag -> 'a
 
 end
 
@@ -154,6 +155,10 @@ module Synterp = struct
   let init_summaries () =
     init_summaries !sum_map_synterp
 
+  (** Summary projection *)
+  let project_from_summary { summaries; _ } tag =
+    Frozen.find tag summaries
+
 end
 
 module Interp = struct
@@ -216,14 +221,4 @@ let ref ?(stage=Stage.Interp) ?(local=false) ~name x =
     in
     r
 
-module Local = struct
-
-type 'a local_ref = 'a ref
-
-let ref ?stage ~name x = ref ?stage ~name ~local:true x
-let (!) = (!)
-let (:=) = (:=)
-
-end
-
 let dump = Dyn.dump
