@@ -10,12 +10,13 @@
 
 module Arg =
 struct
-  module DYN = Dyn.Make(struct end)
+  module DYN = Dyn.Make()
   module Map = DYN.Map
   type ('a, 'b) tag = ('a * 'b) DYN.tag
   let eq = DYN.eq
   let repr = DYN.repr
   let create = DYN.create
+  type glb = Glb : (_,'a) tag * 'a  -> glb
 end
 
 module type Param = sig type ('raw, 'glb) t end
@@ -26,4 +27,4 @@ struct
   include Arg.Map(struct type 'a t = 'a pack end)
 end
 
-module Val = Dyn.Make(struct end)
+module Val = Dyn.Make()
