@@ -25,7 +25,7 @@ let name_table =
 
 type req =
   | ReqLocal
-  | ReqGlobal of GlobRef.t * Name.t list
+  | ReqGlobal
 
 let load_rename_args _ (_, (r, names)) =
   name_table := GlobRef.Map.add r names !name_table
@@ -34,7 +34,7 @@ let cache_rename_args o = load_rename_args 1 o
 
 let classify_rename_args = function
   | ReqLocal, _ -> Dispose
-  | ReqGlobal _, _ -> Substitute
+  | ReqGlobal, _ -> Substitute
 
 let subst_rename_args (subst, (_, (r, names as orig))) =
   ReqLocal,
@@ -42,11 +42,11 @@ let subst_rename_args (subst, (_, (r, names as orig))) =
   if r==r' then orig else (r', names)
 
 let discharge_rename_args = function
-  | ReqGlobal (c, names), _ as req when not (isVarRef c && Lib.is_in_section c) ->
+  | ReqGlobal, (c, names) as req when not (isVarRef c && Lib.is_in_section c) ->
      (try
        let var_names = Array.map_to_list (fun c -> Name (destVar c)) (Lib.section_instance c) in
-       let names' = var_names @ names in
-       Some (ReqGlobal (c, names), (c, names'))
+       let names = var_names @ names in
+       Some (ReqGlobal, (c, names))
      with Not_found -> Some req)
   | _ -> None
 
@@ -70,7 +70,7 @@ let rename_arguments local r names =
             ++ str" may not be renamed.")
     | _ -> ()
   in
-  let req = if local then ReqLocal else ReqGlobal (r, names) in
+  let req = if local then ReqLocal else ReqGlobal in
   Lib.add_leaf (inRenameArgs (req, (r, names)))
 
 let arguments_names r = GlobRef.Map.find r !name_table
