@@ -177,41 +177,66 @@ let safe_gen f env sigma c =
 let safe_pr_lconstr_env = safe_gen pr_lconstr_env
 let safe_pr_constr_env = safe_gen pr_constr_env
 
+let q_ident = Id.of_string "α"
+
 let u_ident = Id.of_string "u"
 
 let universe_binders_with_opt_names orig names =
   let open Univ in
-  let orig = Univ.AbstractContext.names orig in
-  let orig = Array.to_list orig in
-  let udecl = match names with
+  let qorig, uorig = UVars.AbstractContext.names orig in
+  let qorig, uorig as orig = Array.to_list qorig, Array.to_list uorig in
+  let qdecl, udecl = match names with
   | None -> orig
-  | Some udecl ->
+  | Some (qdecl,udecl) ->
     try
-      List.map2 (fun orig {CAst.v = na} ->
-          match na with
-          | Anonymous -> orig
-          | Name id -> Name id) orig udecl
+      let qs =
+        List.map2 (fun orig {CAst.v = na} ->
+            match na with
+            | Anonymous -> orig
+            | Name id -> Name id) qorig qdecl
+      in
+      let us =
+        List.map2 (fun orig {CAst.v = na} ->
+            match na with
+            | Anonymous -> orig
+            | Name id -> Name id) uorig udecl
+      in
+      qs, us
     with Invalid_argument _ ->
       let open UnivGen in
       raise (UniverseLengthMismatch {
-          actual = List.length orig;
-          expect = List.length udecl;
+          actual = List.length qorig, List.length uorig;
+          expect = List.length qdecl, List.length udecl;
         })
   in
-  let fold_named i (ubind,revubind as o) = function
-    | Name id -> let ui = Level.var i in
-      Id.Map.add id ui ubind, Level.Map.add ui id revubind
+  let fold_qnamed i ((qbind,ubind),(revqbind,revubind) as o) = function
+    | Name id -> let ui = Sorts.QVar.make_var i in
+      (Id.Map.add id ui qbind, ubind), (Sorts.QVar.Map.add ui id revqbind, revubind)
     | Anonymous -> o
   in
-  let names = List.fold_left_i fold_named 0 (UnivNames.empty_binders,Level.Map.empty) udecl in
-  let fold_anons i (u_ident, (ubind, revubind) as o) = function
+  let fold_unamed i ((qbind,ubind),(revqbind,revubind) as o) = function
+    | Name id -> let ui = Level.var i in
+      (qbind, Id.Map.add id ui ubind), (revqbind, Level.Map.add ui id revubind)
+    | Anonymous -> o
+  in
+  let names = List.fold_left_i fold_qnamed 0 UnivNames.(empty_binders,empty_rev_binders) qdecl in
+  let names = List.fold_left_i fold_unamed 0 names udecl in
+  let fold_qanons i (u_ident, ((qbind,ubind), (revqbind,revubind)) as o) = function
+    | Name _ -> o
+    | Anonymous ->
+      let ui = Sorts.QVar.make_var i in
+      let id = Namegen.next_ident_away_from u_ident (fun id -> Id.Map.mem id qbind) in
+      (id, ((Id.Map.add id ui qbind, ubind), (Sorts.QVar.Map.add ui id revqbind, revubind)))
+  in
+  let fold_uanons i (u_ident, ((qbind,ubind), (revqbind,revubind)) as o) = function
     | Name _ -> o
     | Anonymous ->
       let ui = Level.var i in
       let id = Namegen.next_ident_away_from u_ident (fun id -> Id.Map.mem id ubind) in
-      (id, (Id.Map.add id ui ubind, Level.Map.add ui id revubind))
+      (id, ((qbind,Id.Map.add id ui ubind), (revqbind,Level.Map.add ui id revubind)))
   in
-  let (_, names) = List.fold_left_i fold_anons 0 (u_ident, names) udecl in
+  let (_, names) = List.fold_left_i fold_qanons 0 (q_ident, names) qdecl in
+  let (_, names) = List.fold_left_i fold_uanons 0 (u_ident, names) udecl in
   names
 
 let pr_universe_ctx_set sigma c =
@@ -221,8 +246,12 @@ let pr_universe_ctx_set sigma c =
     mt()
 
 let pr_universe_ctx sigma ?variance c =
-  if !Detyping.print_universes && not (Univ.UContext.is_empty c) then
-    fnl()++pr_in_comment (v 0 (Univ.pr_universe_context (Termops.pr_evd_level sigma) ?variance c))
+  if !Detyping.print_universes && not (UVars.UContext.is_empty c) then
+    fnl()++
+    pr_in_comment
+      (v 0
+         (UVars.pr_universe_context (Termops.pr_evd_qvar sigma) (Termops.pr_evd_level sigma)
+            ?variance c))
   else
     mt()
 
@@ -230,9 +259,10 @@ let pr_abstract_universe_ctx sigma ?variance ?priv c =
   let open Univ in
   let priv = Option.default Univ.ContextSet.empty priv in
   let has_priv = not (ContextSet.is_empty priv) in
-  if !Detyping.print_universes && (not (Univ.AbstractContext.is_empty c) || has_priv) then
+  if !Detyping.print_universes && (not (UVars.AbstractContext.is_empty c) || has_priv) then
+    let prqvar u = Termops.pr_evd_qvar sigma u in
     let prlev u = Termops.pr_evd_level sigma u in
-    let pub = (if has_priv then str "Public universes:" ++ fnl() else mt()) ++ v 0 (Univ.pr_abstract_universe_context prlev ?variance c) in
+    let pub = (if has_priv then str "Public universes:" ++ fnl() else mt()) ++ v 0 (UVars.pr_abstract_universe_context prqvar prlev ?variance c) in
     let priv = if has_priv then fnl() ++ str "Private universes:" ++ fnl() ++ v 0 (Univ.pr_universe_context_set prlev priv) else mt() in
     fnl()++pr_in_comment (pub ++ priv)
   else
@@ -250,6 +280,7 @@ let pr_global = pr_global_env Id.Set.empty
 
 let pr_universe_instance_constraints evd inst csts =
   let open Univ in
+  let prqvar = Termops.pr_evd_qvar evd in
   let prlev = Termops.pr_evd_level evd in
   let pcsts = if Constraints.is_empty csts then mt()
     else str " |= " ++
@@ -257,7 +288,7 @@ let pr_universe_instance_constraints evd inst csts =
            (fun (u,d,v) -> hov 0 (prlev u ++ pr_constraint_type d ++ prlev v))
            (Constraints.elements csts)
   in
-  str"@{" ++ Instance.pr prlev inst ++ pcsts ++ str"}"
+  str"@{" ++ UVars.Instance.pr prqvar prlev inst ++ pcsts ++ str"}"
 
 let pr_universe_instance evd inst =
   pr_universe_instance_constraints evd inst Univ.Constraints.empty
@@ -281,6 +312,13 @@ let pr_pconstructor = pr_puniverses pr_constructor
 let pr_evaluable_reference ref =
   pr_global (Tacred.global_of_evaluable_reference ref)
 
+let pr_notation_interpretation_env env sigma c =
+  Constrextern.without_symbols (pr_glob_constr_env env sigma) c
+
+let pr_notation_interpretation c =
+  let env = Global.env () in
+  pr_notation_interpretation_env env (Evd.from_env env) c
+
 (*let pr_glob_constr t =
   pr_lconstr (Constrextern.extern_glob_constr Id.Set.empty t)*)
 
@@ -301,17 +339,19 @@ let get_compact_context,set_compact_context =
 let pr_compacted_decl env sigma decl =
   let ids, pbody, typ = match decl with
     | CompactedDecl.LocalAssum (ids, typ) ->
-       ids, mt (), typ
+       ids, None, typ
     | CompactedDecl.LocalDef (ids,c,typ) ->
        (* Force evaluation *)
        let pb = pr_lconstr_env ~inctx:true env sigma c in
        let pb = if isCast c then surround pb else pb in
-       ids, (str" := " ++ pb ++ cut ()), typ
-  in
-  let pids = prlist_with_sep pr_comma (fun id -> pr_id id.binder_name) ids in
+       ids, Some pb, typ in
+  let pids =
+    hov 0 (prlist_with_sep pr_comma (fun id -> pr_id id.binder_name) ids) in
   let pt = pr_ltype_env env sigma typ in
-  let ptyp = (str" : " ++ pt) in
-  hov 0 (pids ++ pbody ++ ptyp)
+  match pbody with
+  | None -> hov 2 (pids ++ str" :" ++ spc () ++ pt)
+  | Some pbody ->
+     hov 2 (pids ++ str" :=" ++ spc () ++ pbody ++ spc () ++ str": " ++ pt)
 
 let pr_ecompacted_decl env sigma (decl:EConstr.compacted_declaration) =
   let Refl = EConstr.Unsafe.eq in
@@ -336,8 +376,8 @@ let pr_rel_decl env sigma decl =
         (str":=" ++ spc () ++ pb ++ spc ()) in
   let ptyp = pr_ltype_env env sigma typ in
   match na with
-  | Anonymous -> hov 0 (str"<>" ++ spc () ++ pbody ++ str":" ++ spc () ++ ptyp)
-  | Name id -> hov 0 (pr_id id ++ spc () ++ pbody ++ str":" ++ spc () ++ ptyp)
+  | Anonymous -> hov 2 (str"<>" ++ spc () ++ pbody ++ str":" ++ spc () ++ ptyp)
+  | Name id -> hov 2 (pr_id id ++ spc () ++ pbody ++ str":" ++ spc () ++ ptyp)
 
 let pr_erel_decl env sigma (decl:EConstr.rel_declaration) =
   let Refl = EConstr.Unsafe.eq in
@@ -364,7 +404,7 @@ let pr_named_context env sigma ne_context =
           ne_context ~init:(mt ()))
 
 let pr_rel_context env sigma rel_context =
-  let rel_context = List.map (fun d -> Termops.map_rel_decl EConstr.of_constr d) rel_context in
+  let rel_context = EConstr.of_rel_context rel_context in
   pr_binders env sigma (extern_rel_context None env sigma rel_context)
 
 let pr_rel_context_of env sigma =
@@ -470,15 +510,18 @@ let pr_predicate pr_elt (b, elts) =
 let pr_cpred p =
   let safe_pr_constant env kn =
     try pr_constant env kn
-    with Not_found when !Flags.in_debugger || !Flags.in_toplevel ->
+    with Not_found when !Flags.in_debugger || !Flags.in_ml_toplevel ->
       Names.Constant.print kn in
   pr_predicate (safe_pr_constant (Global.env())) (Cpred.elements p)
 
 let pr_idpred p = pr_predicate Id.print (Id.Pred.elements p)
 
+let pr_prpred p = pr_predicate Projection.Repr.print (PRpred.elements p)
+
 let pr_transparent_state ts =
   hv 0 (str"VARIABLES: " ++ pr_idpred ts.TransparentState.tr_var ++ fnl () ++
-        str"CONSTANTS: " ++ pr_cpred ts.TransparentState.tr_cst ++ fnl ())
+        str"CONSTANTS: " ++ pr_cpred ts.TransparentState.tr_cst ++ fnl () ++
+        str"PROJECTIONS: " ++ pr_prpred ts.TransparentState.tr_prj ++ fnl ())
 
 let goal_repr sigma g =
   let EvarInfo evi = Evd.find sigma g in
@@ -572,7 +615,7 @@ let pr_evgl_sign env sigma (evi : undefined evar_info) =
 let pr_evar sigma (evk, evi) =
   let env = Global.env () in
   let pegl = pr_evgl_sign env sigma evi in
-  hov 0 (pr_existential_key env sigma evk ++ str " : " ++ pegl)
+  hov 2 (pr_existential_key env sigma evk ++ str " :" ++ spc () ++ pegl)
 
 (* Print an enumerated list of existential variables *)
 let rec pr_evars_int_hd pr sigma i = function
@@ -632,8 +675,8 @@ let print_evar_constraints gl sigma =
     | Some g ->
        let env, _ = goal_repr sigma g in fun e' ->
        begin
-         if Context.Named.equal Constr.equal (named_context env) (named_context e') then
-           if Context.Rel.equal Constr.equal (rel_context env) (rel_context e') then mt ()
+         if Context.Named.equal Sorts.relevance_equal Constr.equal (named_context env) (named_context e') then
+           if Context.Rel.equal Sorts.relevance_equal Constr.equal (rel_context env) (rel_context e') then mt ()
            else pr_rel_context_of e' sigma ++ str " |-" ++ spc ()
          else pr_context_of e' sigma ++ str " |-" ++ spc ()
        end
@@ -701,7 +744,9 @@ let process_dependent_evar q acc evm is_dependent e =
   let () = match Evd.evar_body evi with
   | Evar_empty ->
     queue_term q true (Evd.evar_concl evi)
-  | Evar_defined b -> ()
+  | Evar_defined b ->
+    let env = Evd.evar_filtered_env (Global.env ()) evi in
+    queue_term q true (Retyping.get_type_of env evm b)
   in
   List.iter begin fun decl ->
     let open NamedDecl in
@@ -1031,6 +1076,7 @@ module ContextObjectMap = Map.Make (OrderedContextObject)
 
 let pr_assumptionset env sigma s =
   if ContextObjectMap.is_empty s &&
+       not (rewrite_rules_allowed env) &&
        not (is_impredicative_set env) then
     str "Closed under the global context"
   else
@@ -1054,7 +1100,7 @@ let pr_assumptionset env sigma s =
         MutInd.print kn
     in
     let safe_pr_ltype env sigma typ =
-      try str " : " ++ pr_ltype_env env sigma typ
+      try str " :" ++ spc () ++ pr_ltype_env env sigma typ
       with e when CErrors.noncritical e -> mt ()
     in
     let safe_pr_ltype_relctx (rctx, typ) =
@@ -1065,7 +1111,7 @@ let pr_assumptionset env sigma s =
     let pr_axiom env ax typ =
       match ax with
       | Constant kn ->
-          hov 1 (safe_pr_constant env kn ++ cut() ++ safe_pr_ltype env sigma typ)
+          hov 2 (safe_pr_constant env kn ++ safe_pr_ltype env sigma typ)
       | Positive m ->
           hov 2 (safe_pr_inductive env m ++ spc () ++ strbrk"is assumed to be positive.")
       | Guarded gr ->
@@ -1107,6 +1153,11 @@ let pr_assumptionset env sigma s =
         [str "Set is impredicative"]
       else []
     in
+    let theory =
+      if rewrite_rules_allowed env then
+        str "Rewrite rules are allowed (subject reduction might be broken)" :: theory
+      else theory
+    in
     let theory =
       if type_in_type env then
         str "Type hierarchy is collapsed (logic is inconsistent)" :: theory
