@@ -67,12 +67,12 @@ module RelDecl = Context.Rel.Declaration
 
 let hid = Id.of_string "H"
 let xid = Id.of_string "X"
-let default_id_of_sort = function InSProp | InProp | InSet -> hid | InType -> xid
+let default_id_of_sort = function InSProp | InProp | InSet -> hid | InType | InQSort -> xid
 let fresh env id avoid =
   let freshid = next_global_ident_away id avoid in
   freshid, Id.Set.add freshid avoid
 let with_context_set ctx (b, ctx') =
-  (b, Univ.ContextSet.union ctx ctx')
+  (b, UnivGen.sort_context_union ctx ctx')
 
 let build_dependent_inductive ind (mib,mip) =
   let realargs,_ = List.chop mip.mind_nrealdecls mip.mind_arity_ctxt in
@@ -100,14 +100,10 @@ let my_it_mkLambda_or_LetIn_name env s c =
   List.fold_left (fun c d -> mkLambda_or_LetIn_name d c) c s
 
 let get_coq_eq env ctx =
-  try
-    let eq = Globnames.destIndRef (Coqlib.lib_ref "core.eq.type") in
-    (* Do not force the lazy if they are not defined *)
-    let eq, ctx = with_context_set ctx
+  let eq = Globnames.destIndRef (Coqlib.lib_ref "core.eq.type") in
+  let eq, ctx = with_context_set ctx
       (UnivGen.fresh_inductive_instance env eq) in
-      mkIndU eq, mkConstructUi (eq,1), ctx
-  with Not_found ->
-    user_err Pp.(str "eq not found.")
+  mkIndU eq, mkConstructUi (eq,1), ctx
 
 let univ_of_eq env eq =
   let open EConstr in
@@ -141,7 +137,7 @@ let get_sym_eq_data env (ind,u) =
   if List.exists is_local_def realsign then
     error "Inductive equalities with local definitions in arity not supported.";
   let constrsign,ccl = mip.mind_nf_lc.(0) in
-  let _,constrargs = decompose_app ccl in
+  let _,constrargs = decompose_app_list ccl in
   if not (Int.equal (Context.Rel.length constrsign) (Context.Rel.length mib.mind_params_ctxt)) then
     error "Constructor must have no arguments"; (* This can be relaxed... *)
   let params,constrargs = List.chop mib.mind_nparams constrargs in
@@ -176,7 +172,7 @@ let get_non_sym_eq_data env (ind,u) =
   if List.exists is_local_def realsign then
     error "Inductive equalities with local definitions in arity not supported";
   let constrsign,ccl = mip.mind_nf_lc.(0) in
-  let _,constrargs = decompose_app ccl in
+  let _,constrargs = decompose_app_list ccl in
   if not (Int.equal (Context.Rel.length constrsign) (Context.Rel.length mib.mind_params_ctxt)) then
     error "Constructor must have no arguments";
   let _,constrargs = List.chop mib.mind_nparams constrargs in
@@ -204,31 +200,38 @@ let build_sym_scheme env _handle ind =
     get_sym_eq_data env indu in
   let cstr n =
     mkApp (mkConstructUi(indu,1),Context.Rel.instance mkRel n mib.mind_params_ctxt) in
-  let inds = snd (mind_arity mip) in
+  let inds = Indrec.pseudo_sort_family_for_elim ind mip in
   let varH,_ = fresh env (default_id_of_sort inds) Id.Set.empty in
   let applied_ind = build_dependent_inductive indu specif in
-  let indr = Sorts.relevance_of_sort_family inds in
+  let indr = UVars.subst_instance_relevance u mip.mind_relevance in
   let realsign_ind =
     name_context env ((LocalAssum (make_annot (Name varH) indr,applied_ind))::realsign) in
   let rci = Sorts.Relevant in (* TODO relevance *)
-  let ci = make_case_info env ind rci RegularStyle in
+  let ci = make_case_info env ind RegularStyle in
+  let p =
+    my_it_mkLambda_or_LetIn_name env
+      (lift_rel_context (nrealargs+1) realsign_ind)
+      (mkApp (mkIndU indu,
+              Array.concat
+                [Context.Rel.instance mkRel (3*nrealargs+2) paramsctxt1;
+                 rel_vect 1 nrealargs;
+                 rel_vect (2*nrealargs+2) nrealargs]))
+  in
   let c =
   (my_it_mkLambda_or_LetIn paramsctxt
   (my_it_mkLambda_or_LetIn_name env realsign_ind
-  (mkCase (Inductive.contract_case env (ci,
-     my_it_mkLambda_or_LetIn_name env
-       (lift_rel_context (nrealargs+1) realsign_ind)
-       (mkApp (mkIndU indu,Array.concat
-          [Context.Rel.instance mkRel (3*nrealargs+2) paramsctxt1;
-           rel_vect 1 nrealargs;
-           rel_vect (2*nrealargs+2) nrealargs])),
-           NoInvert,
-     mkRel 1 (* varH *),
-       [|cstr (nrealargs+1)|])))))
-  in c, UState.of_context_set ctx
+     (mkCase
+        (Inductive.contract_case env
+           (ci,
+            (p,rci),
+            NoInvert,
+            mkRel 1 (* varH *),
+            [|cstr (nrealargs+1)|])))))
+  in
+  c, UState.of_context_set ctx
 
 let sym_scheme_kind =
-  declare_individual_scheme_object "_sym_internal"
+  declare_individual_scheme_object "sym_internal"
   build_sym_scheme
 
 (**********************************************************************)
@@ -260,8 +263,8 @@ let build_sym_involutive_scheme env handle ind =
   let eq,eqrefl,ctx = get_coq_eq env ctx in
   let sym, ctx = const_of_scheme sym_scheme_kind env handle ind ctx in
   let cstr n = mkApp (mkConstructUi (indu,1),Context.Rel.instance mkRel n paramsctxt) in
-  let inds = snd (mind_arity mip) in
-  let indr = Sorts.relevance_of_sort_family inds in
+  let inds = Indrec.pseudo_sort_family_for_elim ind mip in
+  let indr = UVars.subst_instance_relevance u mip.mind_relevance in
   let varH,_ = fresh env (default_id_of_sort inds) Id.Set.empty in
   let applied_ind = build_dependent_inductive indu specif in
   let applied_ind_C =
@@ -272,12 +275,12 @@ let build_sym_involutive_scheme env handle ind =
   let realsign_ind =
     name_context env ((LocalAssum (make_annot (Name varH) indr,applied_ind))::realsign) in
   let rci = Sorts.Relevant in (* TODO relevance *)
-  let ci = make_case_info env ind rci RegularStyle in
+  let ci = make_case_info env ind RegularStyle in
   let c =
     (my_it_mkLambda_or_LetIn paramsctxt
      (my_it_mkLambda_or_LetIn_name env realsign_ind
       (mkCase (Inductive.contract_case env (ci,
-                my_it_mkLambda_or_LetIn_name env
+                (my_it_mkLambda_or_LetIn_name env
                 (lift_rel_context (nrealargs+1) realsign_ind)
                 (mkApp (eq,[|
                 mkApp
@@ -294,14 +297,14 @@ let build_sym_involutive_scheme env handle ind =
                  rel_vect (2*nrealargs+2) nrealargs;
                  rel_vect 1 nrealargs;
                  [|mkRel 1|]])|]]);
-               mkRel 1|])),
+               mkRel 1|])), rci),
                NoInvert,
                mkRel 1 (* varH *),
                [|mkApp(eqrefl,[|applied_ind_C;cstr (nrealargs+1)|])|])))))
   in (c, UState.of_context_set ctx)
 
 let sym_involutive_scheme_kind =
-  declare_individual_scheme_object "_sym_involutive"
+  declare_individual_scheme_object "sym_involutive"
   ~deps:(fun _ ind -> [SchemeIndividualDep (ind, sym_scheme_kind)])
   build_sym_involutive_scheme
 
@@ -376,8 +379,8 @@ let build_l2r_rew_scheme dep env handle ind kind =
     mkApp (mkConstructUi(indu,1),
       Array.concat [Context.Rel.instance mkRel n paramsctxt1;
                     rel_vect p nrealargs]) in
-  let inds = snd (mind_arity mip) in
-  let indr = Sorts.relevance_of_sort_family inds in
+  let inds = Indrec.pseudo_sort_family_for_elim ind mip in
+  let indr = UVars.subst_instance_relevance u mip.mind_relevance in
   let varH,avoid = fresh env (default_id_of_sort inds) Id.Set.empty in
   let varHC,avoid = fresh env (Id.of_string "HC") avoid in
   let varP,_ = fresh env (Id.of_string "P") avoid in
@@ -408,11 +411,11 @@ let build_l2r_rew_scheme dep env handle ind kind =
                      rel_vect 1 nrealargs;
                      [|mkRel 1|]]) in
   let s, ctx' = UnivGen.fresh_sort_in_family kind in
-  let ctx = Univ.ContextSet.union ctx ctx' in
+  let ctx = UnivGen.sort_context_union ctx ctx' in
   let s = mkSort s in
   let rci = Sorts.Relevant in (* TODO relevance *)
-  let ci = make_case_info env ind rci RegularStyle in
-  let cieq = make_case_info env (fst (destInd eq)) rci RegularStyle in
+  let ci = make_case_info env ind RegularStyle in
+  let cieq = make_case_info env (fst (destInd eq)) RegularStyle in
   let applied_PC =
     mkApp (mkVar varP,Array.append (Context.Rel.instance mkRel 1 realsign)
            (if dep then [|cstr (2*nrealargs+1) 1|] else [||])) in
@@ -434,7 +437,7 @@ let build_l2r_rew_scheme dep env handle ind kind =
                [|mkRel 2|]])|]]) in
   let main_body =
     mkCase (Inductive.contract_case env (ci,
-            my_it_mkLambda_or_LetIn_name env realsign_ind_G applied_PG,
+            (my_it_mkLambda_or_LetIn_name env realsign_ind_G applied_PG, rci),
             NoInvert,
             applied_sym_C 3,
             [|mkVar varHC|]))
@@ -448,10 +451,10 @@ let build_l2r_rew_scheme dep env handle ind kind =
   (mkNamedLambda (make_annot varH indr) (lift 2 applied_ind)
      (if dep then (* we need a coercion *)
      mkCase (Inductive.contract_case env (cieq,
-       mkLambda (make_annot (Name varH) indr,lift 3 applied_ind,
+       (mkLambda (make_annot (Name varH) indr,lift 3 applied_ind,
          mkLambda (make_annot Anonymous indr,
                    mkApp (eq,[|lift 4 applied_ind;applied_sym_sym;mkRel 1|]),
-                   applied_PR)),
+                   applied_PR)), rci),
        NoInvert,
        mkApp (sym_involutive,
          Array.append (Context.Rel.instance mkRel 3 mip.mind_arity_ctxt) [|mkVar varH|]),
@@ -494,8 +497,8 @@ let build_l2r_forward_rew_scheme dep env ind kind =
     mkApp (mkConstructUi(indu,1),
       Array.concat [Context.Rel.instance mkRel n paramsctxt1;
                     rel_vect p nrealargs]) in
-  let inds = snd (mind_arity mip) in
-  let indr = Sorts.relevance_of_sort_family inds in
+  let inds = Indrec.pseudo_sort_family_for_elim ind mip in
+  let indr = UVars.subst_instance_relevance u mip.mind_relevance in
   let varH,avoid = fresh env (default_id_of_sort inds) Id.Set.empty in
   let varHC,avoid = fresh env (Id.of_string "HC") avoid in
   let varP,_ = fresh env (Id.of_string "P") avoid in
@@ -516,10 +519,10 @@ let build_l2r_forward_rew_scheme dep env ind kind =
   let realsign_ind_P n aP =
     name_context env ((LocalAssum (make_annot (Name varH) indr,aP))::realsign_P n) in
   let s, ctx' = UnivGen.fresh_sort_in_family kind in
-  let ctx = Univ.ContextSet.union ctx ctx' in
+  let ctx = UnivGen.sort_context_union ctx ctx' in
   let s = mkSort s in
   let rci = Sorts.Relevant in
-  let ci = make_case_info env ind rci RegularStyle in
+  let ci = make_case_info env ind RegularStyle in
   let applied_PC =
     mkApp (mkVar varP,Array.append
            (rel_vect (nrealargs*2+3) nrealargs)
@@ -537,12 +540,12 @@ let build_l2r_forward_rew_scheme dep env ind kind =
   (my_it_mkLambda_or_LetIn_name env realsign
   (mkNamedLambda (make_annot varH indr) applied_ind
   (mkCase (Inductive.contract_case env (ci,
-     my_it_mkLambda_or_LetIn_name env
+     (my_it_mkLambda_or_LetIn_name env
        (lift_rel_context (nrealargs+1) realsign_ind)
        (mkNamedProd (make_annot varP indr)
          (my_it_mkProd_or_LetIn
            (if dep then realsign_ind_P 2 applied_ind_P else realsign_P 2) s)
-       (mkNamedProd (make_annot varHC indr) applied_PC applied_PG)),
+       (mkNamedProd (make_annot varHC indr) applied_PC applied_PG)), rci),
      NoInvert,
      (mkVar varH),
      [|mkNamedLambda (make_annot varP indr)
@@ -589,8 +592,8 @@ let build_r2l_forward_rew_scheme dep env ind kind =
   let cstr n =
     mkApp (mkConstructUi(indu,1),Context.Rel.instance mkRel n mib.mind_params_ctxt) in
   let constrargs_cstr = constrargs@[cstr 0] in
-  let inds = snd (mind_arity mip) in
-  let indr = Sorts.relevance_of_sort_family inds in
+  let inds = Indrec.pseudo_sort_family_for_elim ind mip in
+  let indr = Inductive.relevance_of_ind_body mip u in
   let varH,avoid = fresh env (default_id_of_sort inds) Id.Set.empty in
   let varHC,avoid = fresh env (Id.of_string "HC") avoid in
   let varP,_ = fresh env (Id.of_string "P") avoid in
@@ -598,10 +601,10 @@ let build_r2l_forward_rew_scheme dep env ind kind =
   let realsign_ind =
     name_context env ((LocalAssum (make_annot (Name varH) indr,applied_ind))::realsign) in
   let s, ctx' = UnivGen.fresh_sort_in_family kind in
-  let ctx = Univ.ContextSet.union ctx ctx' in
+  let ctx = UnivGen.sort_context_union ctx ctx' in
   let s = mkSort s in
   let rci = Sorts.Relevant in (* TODO relevance *)
-  let ci = make_case_info env ind rci RegularStyle in
+  let ci = make_case_info env ind RegularStyle in
   let applied_PC =
     applist (mkVar varP,if dep then constrargs_cstr else constrargs) in
   let applied_PG =
@@ -617,9 +620,9 @@ let build_r2l_forward_rew_scheme dep env ind kind =
   (mkNamedLambda (make_annot varHC indr) (lift 1 applied_PG)
   (mkApp
     (mkCase (Inductive.contract_case env (ci,
-       my_it_mkLambda_or_LetIn_name env
+       (my_it_mkLambda_or_LetIn_name env
          (lift_rel_context (nrealargs+3) realsign_ind)
-         (mkArrow applied_PG indr (lift (2*nrealargs+5) applied_PC)),
+         (mkArrow applied_PG indr (lift (2*nrealargs+5) applied_PC)), rci),
        NoInvert,
        mkRel 3 (* varH *),
        [|mkLambda
@@ -649,7 +652,7 @@ let fix_r2l_forward_rew_scheme env (c, ctx') =
   let sigma = Evd.from_env env in
   let t = Retyping.get_type_of env sigma (EConstr.of_constr c) in
   let t = EConstr.Unsafe.to_constr t in
-  let ctx,_ = decompose_prod_assum t in
+  let ctx,_ = decompose_prod_decls t in
   match ctx with
   | hp :: p :: ind :: indargs ->
      let c' =
@@ -687,8 +690,11 @@ let fix_r2l_forward_rew_scheme env (c, ctx') =
 let build_r2l_rew_scheme dep env ind k =
   let sigma = Evd.from_env env in
   let (sigma, indu) = Evd.fresh_inductive_instance env sigma ind in
-  let (sigma, c, _) = build_case_analysis_scheme env sigma indu dep k in
-    c, Evd.evar_universe_context sigma
+  let indu = on_snd EConstr.EInstance.make indu in
+  let sigma, k = Evd.fresh_sort_in_family ~rigid:UnivRigid sigma k in
+  let (sigma, c) = build_case_analysis_scheme env sigma indu dep k in
+  let (c, _) = Indrec.eval_case_analysis c in
+  EConstr.Unsafe.to_constr c, Evd.evar_universe_context sigma
 
 (**********************************************************************)
 (* Register the rewriting schemes                                     *)
@@ -701,7 +707,7 @@ let build_r2l_rew_scheme dep env ind k =
 (* with H:I p1..pn a1..an in Gamma                                    *)
 (**********************************************************************)
 let rew_l2r_dep_scheme_kind =
-  declare_individual_scheme_object "_rew_r_dep"
+  declare_individual_scheme_object "rew_r_dep"
   ~deps:(fun _ ind -> [
     SchemeIndividualDep (ind, sym_scheme_kind);
     SchemeIndividualDep (ind, sym_involutive_scheme_kind);
@@ -715,7 +721,7 @@ let rew_l2r_dep_scheme_kind =
 (* or   H:I b1..bn a1..an in Gamma (symmetric case)                   *)
 (**********************************************************************)
 let rew_r2l_dep_scheme_kind =
-  declare_individual_scheme_object "_rew_dep"
+  declare_individual_scheme_object "rew_dep"
   (fun env _ ind -> build_r2l_rew_scheme true env ind InType)
 
 (**********************************************************************)
@@ -725,7 +731,7 @@ let rew_r2l_dep_scheme_kind =
 (* or   H:I b1..bn a1..an in Gamma (symmetric case)                   *)
 (**********************************************************************)
 let rew_r2l_forward_dep_scheme_kind =
-  declare_individual_scheme_object "_rew_fwd_dep"
+  declare_individual_scheme_object "rew_fwd_dep"
   (fun env _ ind -> build_r2l_forward_rew_scheme true env ind InType)
 
 (**********************************************************************)
@@ -735,7 +741,7 @@ let rew_r2l_forward_dep_scheme_kind =
 (* with H:I p1..pn a1..an in Gamma                                    *)
 (**********************************************************************)
 let rew_l2r_forward_dep_scheme_kind =
-  declare_individual_scheme_object "_rew_fwd_r_dep"
+  declare_individual_scheme_object "rew_fwd_r_dep"
   (fun env _ ind -> build_l2r_forward_rew_scheme true env ind InType)
 
 (**********************************************************************)
@@ -748,7 +754,7 @@ let rew_l2r_forward_dep_scheme_kind =
 (* standard form of schemes in Coq)                                   *)
 (**********************************************************************)
 let rew_l2r_scheme_kind =
-  declare_individual_scheme_object "_rew_r"
+  declare_individual_scheme_object "rew_r"
   (fun env _ ind -> fix_r2l_forward_rew_scheme env
      (build_r2l_forward_rew_scheme false env ind InType))
 
@@ -759,7 +765,7 @@ let rew_l2r_scheme_kind =
 (* introducing commutative cuts, we adopt it                          *)
 (**********************************************************************)
 let rew_r2l_scheme_kind =
-  declare_individual_scheme_object "_rew"
+  declare_individual_scheme_object "rew"
   (fun env _ ind -> build_r2l_rew_scheme false env ind InType)
 
 (* End of rewriting schemes *)
@@ -799,7 +805,7 @@ let build_congr env (eq,refl,ctx) ind =
     RelDecl.get_type decl, RelDecl.get_relevance decl
   in
   let constrsign,ccl = mip.mind_nf_lc.(0) in
-  let _,constrargs = decompose_app ccl in
+  let _,constrargs = decompose_app_list ccl in
   if not (Int.equal (Context.Rel.length constrsign) (Context.Rel.length mib.mind_params_ctxt)) then
     error "Constructor must have no arguments";
   let b = List.nth constrargs (i + mib.mind_nparams - 1) in
@@ -807,9 +813,12 @@ let build_congr env (eq,refl,ctx) ind =
   let varH,avoid = fresh env (Id.of_string "H") avoid in
   let varf,avoid = fresh env (Id.of_string "f") avoid in
   let rci = Sorts.Relevant in (* TODO relevance *)
-  let ci = make_case_info env ind rci RegularStyle in
-  let uni, ctx = Univ.extend_in_context_set (UnivGen.new_global_univ ()) ctx in
-  let ctx = (fst ctx, UnivSubst.enforce_leq uni (univ_of_eq env eq) (snd ctx)) in
+  let ci = make_case_info env ind RegularStyle in
+  let uni, ctx' = UnivGen.new_global_univ () in
+  let ctx =
+    let (qs,us),csts = ctx in
+    let us, csts = Univ.ContextSet.union (us,csts) ctx' in
+    ((qs, us), UnivSubst.enforce_leq uni (univ_of_eq env eq) csts) in
   let c =
   my_it_mkLambda_or_LetIn paramsctxt
      (mkNamedLambda (make_annot varB Sorts.Relevant) (mkType uni)
@@ -821,7 +830,7 @@ let build_congr env (eq,refl,ctx) ind =
             Context.Rel.instance_list mkRel (mip.mind_nrealargs+2) paramsctxt @
             Context.Rel.instance_list mkRel 0 realsign))
      (mkCase (Inductive.contract_case env (ci,
-       my_it_mkLambda_or_LetIn_name env
+       (my_it_mkLambda_or_LetIn_name env
          (lift_rel_context (mip.mind_nrealargs+3) realsign)
          (mkLambda
            (make_annot Anonymous Sorts.Relevant,
@@ -833,7 +842,7 @@ let build_congr env (eq,refl,ctx) ind =
             mkApp (eq,
               [|mkVar varB;
                 mkApp (mkVar varf, [|lift (2*mip.mind_nrealdecls+4) b|]);
-                mkApp (mkVar varf, [|mkRel (mip.mind_nrealargs - i + 2)|])|]))),
+                mkApp (mkVar varf, [|mkRel (mip.mind_nrealargs - i + 2)|])|]))), rci),
        NoInvert,
        mkVar varH,
        [|mkApp (refl,
@@ -841,7 +850,7 @@ let build_congr env (eq,refl,ctx) ind =
             mkApp (mkVar varf, [|lift (mip.mind_nrealargs+3) b|])|])|])))))))
   in c, UState.of_context_set ctx
 
-let congr_scheme_kind = declare_individual_scheme_object "_congr"
+let congr_scheme_kind = declare_individual_scheme_object "congr"
   (fun env _ ind ->
      (* May fail if equality is not defined *)
-   build_congr env (get_coq_eq env Univ.ContextSet.empty) ind)
+   build_congr env (get_coq_eq env UnivGen.empty_sort_context) ind)
