@@ -7,9 +7,9 @@
 dotest=true
 if [ $dotest = false ]; then exit 0; fi
 rm -f misc/deps/Theory2/*.vo misc/deps/Theory2/Subtheory?/*.vo misc/deps/Theory2/Subtheory?/Subsubtheory?/*.vo
-tmpoutput=$(mktemp /tmp/coqcheck.XXXXXX)
-(cd misc/deps; $coqdep -f _CoqTheory2Project) > "$tmpoutput" 2>&1
-diff -u --strip-trailing-cr misc/deps/Theory2Deps.out $tmpoutput
+output=misc/deps/Theory2Deps.real
+(cd misc/deps; $coqdep -f _CoqTheory2Project) > "$output" 2>&1
+diff -u --strip-trailing-cr misc/deps/Theory2Deps.out $output
 R=$?
 if [ $R != 0 ]; then
     printf "Unexpected coqdep result.\n"
