@@ -2,6 +2,7 @@ open Names
 
 let evil name name_f =
   let open Univ in
+  let open UVars in
   let open Constr in
   let kind = Decls.(IsDefinition Definition) in
   let u = Level.var 0 in
@@ -13,7 +14,7 @@ let evil name name_f =
   let tc = mkConst tc in
 
   let fe = Declare.definition_entry
-      ~univs:(UState.Polymorphic_entry (UContext.make [|Anonymous|] (Instance.of_array [|u|],Constraints.empty)), UnivNames.empty_binders)
+      ~univs:(UState.Polymorphic_entry (UContext.make ([||],[|Anonymous|]) (Instance.of_array ([||],[|u|]),Constraints.empty)), UnivNames.empty_binders)
       ~types:(Term.mkArrowR tc tu)
       (mkLambda (Context.nameR (Id.of_string "x"), tc, mkRel 1))
   in
