@@ -21,11 +21,8 @@ module Rule_type : sig
   type native = Disabled | Coqc | CoqNative
   type t =
     | Regular of { native : native }
-    | Quick
-    (** build / depend on .vio instead of .vo  *)
 
   val native_coqc : t -> bool
-  val vo_ext : t -> string
 end
 
 (** Return the native object files for a module *)
@@ -46,3 +43,5 @@ val install_files :
   -> (string * string) list
 
 val pp : Format.formatter -> t -> unit
+
+val with_timing : bool
