@@ -127,7 +127,7 @@ let init_document opts stm_options injections =
            { doc_type = Interactive opts.config.logic.toplevel_name;
              injections;
            }) in
-  { doc; sid; proof = None; time = opts.config.time }
+  { doc; sid; proof = None; time = Option.map Vernac.make_time_output opts.config.time }
 
 let init_toploop opts stm_opts injections =
   let state = init_document opts stm_opts injections in
@@ -173,7 +173,7 @@ let get_native_name s =
 
 let coqtop_run ({ run_mode; color_mode },_) ~opts state =
   match run_mode with
-  | Interactive -> Coqloop.loop ~opts ~state;
+  | Interactive -> Coqloop.run ~opts ~state;
   | Query PrintTags -> Colors.print_style_tags color_mode; exit 0
   | Query (PrintModUid sl) ->
       let s = String.concat " " (List.map get_native_name sl) in
